package edu.cmu.cs.sb.joint_sdrem;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Date;
import java.util.LinkedList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.Random;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Properties;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;
import java.io.PrintWriter;

import edu.cmu.cs.sb.drem.DREMGui_SaveModel;
import edu.cmu.cs.sb.drem.DREM_Timeiohmm;
import edu.cmu.cs.sb.drem.DREM_IO_Batch;
import gnu.trove.map.hash.*;
import gnu.trove.set.hash.*;
import gnu.trove.iterator.*;
import gnu.trove.iterator.hash.*;
import gnu.trove.list.array.TIntArrayList;
import gnu.trove.list.array.TDoubleArrayList;
import java.io.*;
import util.ArrayMapInt;
import util.ArraySet;

import ilog.concert.*;
import ilog.cplex.*;


public class JointSDREM {
	
	public static class NetworkData {
		int numEdges;
		int[][] edges;
		int[][] edgeToNodes;
		TIntIntHashMap [] edgeToId;
		byte[] edgeType; // 0 = pp, 1 = ptm, 2 = pd
		double[] edgeCost;
		double[] nodePriors;
		int numNodes;
		HashMap<String, Integer> geneNameToId = new HashMap<String, Integer>();

		TIntHashSet [] sourceSet;

		// Target to target Index
		ArrayMapInt allTargets;
		int numAllTargets;

		int [] targetIndexToTarget;
		String [] geneIdToName = null;
		
		public NetworkData() {
			geneNameToId = new HashMap<String, Integer>();
		}
	}
	
	public static class Params {
		double convergenceRatio;
		String edgeFile;
		String [] srcFiles;
		String dataDir = "";
		String modelDir;
		String dremSettingsFile;
		String tfGenePriorsFile;
		String nodePriorsFile;
		String storedPathsDir;
		int iterations;
		int pathLength;
		int pathEnumBound;
		int orientationRestarts;
		int dremRandomizations;
		int distTfs;
		int topRankedPaths;
		int tabuSize;
		int tabuMaxItr;
		int[] cond_to_class;
		int numClasses;
		//int[][] class_to_cond;
		double distThresh;
		double targetThresh;
		double nodeThresh;
		double defaultNodePrior;
		double minPrior;
		float randTargThreshold;
		int startIteration;
		long seed;
		int numConditions;
		double defaultTFprior;
		boolean useWarmDremStart;
		double tfShareFactor;
		
		double maxFlowSourcePenalty;
		String networkMethod;
		
		public Params(Properties props) {
			edgeFile = props.getProperty("edges.file");
			String srcFilesString = props.getProperty("sources.file");
			assert srcFilesString != "" : "srcFilesString empty...";
			srcFiles = srcFilesString.split("\t");
			numConditions = srcFiles.length;
			assert numConditions > 0: "numConditions is becoming 0... with string " + srcFilesString;

			String cond_class_str = props.getProperty("cond.class");
			if (cond_class_str == "") {
				cond_to_class = new int[numConditions];
				numClasses = 1;
				/*
				class_to_cond = new int[1][numConditions];
				for (int i=0; i<numConditions; ++i) {
					class_to_cond[0][i] = i;
				}
				*/
			}
			else {
				String[] cond_class_split = cond_class_str.split("\t");
				assert cond_class_split.length == srcFiles.length : cond_class_split.length + " == " + srcFiles.length;
				cond_to_class = new int[cond_class_split.length];
				HashMap<String, Integer> class_to_id = new HashMap<String, Integer>();
				//TIntIntHashMap class_to_num = new TIntIntHashMap();
				for (int i=0; i<cond_class_split.length; ++i) {
					String c = cond_class_split[i].trim();
					if (!class_to_id.containsKey(c)) {
						int s = class_to_id.size();
						class_to_id.put(c, s);
						//class_to_num.put(s, 0);
					}
	
					cond_to_class[i] = class_to_id.get(c);
					//class_to_num.increment(cond_to_class[i]);
				}
				
				numClasses = class_to_id.size();

				/*
				class_to_cond = new int[class_to_id.size()][];
				for (int i=0; i<class_to_num.size(); ++i) {
					class_to_cond[i] = new int[class_to_num.get(i)];
				}
				for (int i=0; i<cond_to_class.length; ++i) {
					int k = class_to_num.get(cond_to_class[i])-1;
					class_to_cond[cond_to_class[i]][k] = i;
					class_to_num.adjustValue(i, k);
				}
				*/
			}

			modelDir = props.getProperty("model.dir");
			dremSettingsFile = props.getProperty("drem.settings.file");
			tfGenePriorsFile = props.getProperty("binding.priors.file");
			nodePriorsFile = props.getProperty("node.priors.file");
			storedPathsDir = props.getProperty("stored.paths.dir");
			iterations = Integer.parseInt(props.getProperty("iterations"));
			pathLength = Integer.parseInt(props.getProperty("max.path.length"));
			pathEnumBound = Integer.parseInt(props.getProperty("path.enum.bound"));
			orientationRestarts = Integer.parseInt(props.getProperty("orientation.restarts"));
			dremRandomizations = Integer.parseInt(props.getProperty("drem.random.runs"));
			distTfs = Integer.parseInt(props.getProperty("dist.tfs"));
			topRankedPaths = Integer.parseInt(props.getProperty("top.paths"));
			distThresh = Double.parseDouble(props.getProperty("dist.thresh"));
			targetThresh = Double.parseDouble(props.getProperty("target.thresh"));
			nodeThresh = Double.parseDouble(props.getProperty("node.thresh"));
			defaultNodePrior = Double.parseDouble(props.getProperty("default.node.prior"));
			minPrior = Double.parseDouble(props.getProperty("min.prior"));
			randTargThreshold = Float.parseFloat(props.getProperty("random.target.ratio"));
			startIteration = Integer.parseInt(props.getProperty("startIteration"));
			seed = Long.parseLong(props.getProperty("seed"));
			tabuSize = Integer.parseInt(props.getProperty("tabu.size"));
			tabuMaxItr = Integer.parseInt(props.getProperty("tabu.maxitr"));
			defaultTFprior = Double.parseDouble(props.getProperty("default.TF.prior"));
			DREMInterface.defaultTFprior = defaultTFprior;
			DREM_Timeiohmm.defaultTFprior = defaultTFprior;
			DREMGui_SaveModel.SYN_FILE = props.getProperty("synFile");
			Global.DEBUG = Boolean.parseBoolean(props.getProperty("DEBUG"));
			Global.XDEBUG = Boolean.parseBoolean(props.getProperty("XDEBUG"));
			convergenceRatio = Double.parseDouble(props.getProperty("opt.convergence.ratio"));
			useWarmDremStart = Boolean.parseBoolean(props.getProperty("WarmDremStart"));
			tfShareFactor = Double.parseDouble(props.getProperty("tf.share.factor"));
			
			// Options - path, maxflow, maxflow.orientation
			maxFlowSourcePenalty = Double.parseDouble(props.getProperty("maxflow.source.penalty"));
			networkMethod = props.getProperty("network.method");
		}
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args) throws Exception {
		if(args.length < 1)
		{
			throw new RuntimeException("Not enough parameters\n" +
				"Usage: Joint_SDREM <properties_file>\n");
		}
		
		runSDREM(args[0]);
	}
	
	public static boolean isPathActive(
			final int[] path,
			final byte[] edgeDir)
	{
		for (int j=0; j<path.length; ++j) {
			if (edgeDir[path[j]] == 0)
				return false;
		}
		
		return true;
	}

	public static void flip(
			final int flipEdge,
			byte[] edgeDir,
			boolean[] pathActive,
			double[] flipCost,
			double [] pathCost,
			final TIntHashSet targets,
			final PathData pathData,
			final NetworkData networkData)
	{
		assert flipEdge < networkData.numEdges;
		assert networkData.edgeType[flipEdge] == (byte) 0 : flipEdge + " type is " 
				+ networkData.edgeType[flipEdge];
		edgeDir[flipEdge] = (byte) (1-edgeDir[flipEdge]);
		edgeDir[flipEdge+networkData.numEdges] = (byte) (1-edgeDir[flipEdge]);
		final int[][] paths = pathData.paths;
		final int[][] undirPaths = pathData.undirPaths;
		final int [][][] edgeToPaths = pathData.edgeToPaths;

		flipCost[flipEdge] = 0;
		for (int i=0; i<edgeToPaths[flipEdge].length; ++i) {
			int path_id = edgeToPaths[flipEdge][i][0];
			int[] path = paths[path_id];
			int tf = networkData.edgeToNodes[path[path.length-1]][1];
			assert networkData.allTargets.getSet().contains(tf);
			if (!targets.contains(tf)) {
				assert !pathActive[path_id];
				continue;
			}
			int flipEdgeIndex = edgeToPaths[flipEdge][i][1];
			
			if (!pathActive[path_id]) {
				pathActive[path_id] = isPathActive(path, edgeDir);
				if (pathActive[path_id]) {
					for (int j=0; j<path.length; ++j) {
						flipCost[undirPaths[path_id][j]] -= pathCost[path_id];
					}
				}
				else {
					if (edgeDir[path[flipEdgeIndex]] == 0) {
						int lonerEdge = -1;
						for (int j=0; j<path.length; ++j) {
							if (undirPaths[path_id][j] == flipEdge) {
								continue;
							}

							if (lonerEdge != -1 && edgeDir[path[j]] == 0) {
								lonerEdge = -1;
								break;
							}
							else if (lonerEdge == -1 && edgeDir[path[j]] == 0) {
								lonerEdge = undirPaths[path_id][j];
							}
						}
						if (lonerEdge != -1)
							flipCost[lonerEdge] -= pathCost[path_id];
					}
					else {
						int lonerEdge = -1;
						for (int j=0; j<path.length; ++j) {
							if (lonerEdge != -1 && edgeDir[path[j]] == 0) {
								lonerEdge = -1;
								break;
							}
							else if (lonerEdge == -1 && edgeDir[path[j]] == 0) {
								lonerEdge = undirPaths[path_id][j];
							}
						}
						if (lonerEdge != -1)
							flipCost[lonerEdge] += pathCost[path_id];
					}
				}
			}
			else {
				// In this case, flipping even one edge direction eliminates the path
				pathActive[path_id] = false;
				for (int j=0; j<path.length; ++j) {
					flipCost[undirPaths[path_id][j]] += pathCost[path_id];
				}
			}
		}
	}
	
	public static void computeAllFlipCosts(
			final byte[] edgeDir,
			double[] flipCost,
			boolean[] pathActive,
			final double[] pathCost,
			final int[][] paths,
			final int[][] undirPaths)
	{

		for (int i=0; i<flipCost.length; ++i) {
			flipCost[i] = 0.;
		}
		
		for (int i=0; i<paths.length; ++i) {
			if (pathActive[i]) {
				for (int j=0; j<paths[i].length; ++j) {
					flipCost[undirPaths[i][j]] -= pathCost[i];
				}
			}
			else {
				int lonerEdge = -1;
				for (int j=0; j<paths[i].length; ++j) {
					if (lonerEdge != -1 && edgeDir[paths[i][j]] == 0) {
						lonerEdge = -1;
						break;
					}
					else if (lonerEdge == -1 && edgeDir[paths[i][j]] == 0) {
						lonerEdge = undirPaths[i][j];
					}
				}
				if (lonerEdge != -1)
					flipCost[lonerEdge] += pathCost[i];
			}
		}
	}
	
	public static String edgeToString(int[] nodes, String[] geneIdToName) {
		return geneIdToName[nodes[0]] + "\t" + geneIdToName[nodes[1]];
	}
	
	public static boolean [] randomFlipping(
			byte[] edgeDir,
			double[] flipCost,
			boolean[] pathActive,
			double[] pathCost,
			PathData pathData,
			final NetworkData networkData,
			final TIntHashSet targets,
			Random r)
	{
		final int[][] paths = pathData.paths;
		final int[][] undirPaths = pathData.undirPaths;
		int numEdges = networkData.numEdges;
		assert numEdges > 0;
		
		boolean [] fixedEdges = new boolean[numEdges];
		int[] edgeInPath = new int[numEdges];
		for (int i=0; i<fixedEdges.length; ++i) {
			fixedEdges[i] = true;
			edgeInPath[i] = -1;
		}

		for (int i=0; i<paths.length; ++i) {
			int tf = networkData.edgeToNodes[paths[i][paths[i].length-1]][1];
			if (!targets.contains(tf)) {
				continue;
			}
			for (int j=0; j<paths[i].length; ++j) {
				int undirEdge = undirPaths[i][j];
				if (!fixedEdges[undirEdge])
					continue;
				if (edgeInPath[undirEdge] != -1 && edgeInPath[undirEdge] != paths[i][j]) {
					fixedEdges[undirEdge] = false;
				}
				else {
					edgeInPath[undirEdge] = paths[i][j];
				}
			}
		}
		
		int numFixed = 0;
		for (int i=0; i<numEdges; ++i) {
			if (fixedEdges[i])
				++numFixed;

			if (networkData.edgeType[i] != 0) {
				// TODO assert fixedEdges[i] should work here but is not. Temp fix.
				assert fixedEdges[i] : edgeToString(networkData.edgeToNodes[i], networkData.geneIdToName) + "\t" + i
				+ "\t" + networkData.edgeToId[networkData.edgeToNodes[i][0]].get(networkData.edgeToNodes[i][1]);
				edgeDir[i] = 1;
				edgeDir[i+networkData.numEdges] = 0;
			}
			else if (fixedEdges[i]) {
				if (edgeInPath[i] == -1) {
					edgeDir[i] = 1;
					edgeDir[i+numEdges] = 1;
				}
				else {
					edgeDir[edgeInPath[i]] = 1;
					if (edgeInPath[i] >= numEdges) {
						edgeDir[edgeInPath[i]-numEdges] = 0;
					}
					else {
						edgeDir[edgeInPath[i]+numEdges] = 0;
					}
				}
			}
			else {
				edgeDir[i] = (byte) r.nextInt(2);
				edgeDir[i+networkData.numEdges] = (byte)(1-edgeDir[i]);
			}
		}
		for (int i=0; i<paths.length; ++i) {
			int tf = networkData.edgeToNodes[paths[i][paths[i].length-1]][1];
			if (!targets.contains(tf)) {
				pathActive[i] = false;
				continue;
			}
			pathActive[i] = isPathActive(paths[i], edgeDir);
		}
		
		System.out.println("Done random flipping with " + numFixed + " fixed edges out of " + numEdges + " total edges for " + (numEdges-numFixed) + " free edges");
		
		computeAllFlipCosts(edgeDir, flipCost, pathActive, pathCost, paths, undirPaths);
		return fixedEdges;
	}
	
	public static void readTargets(
			String dremTargetsFile,
			ArraySet[] realTargets,
			double [][] realTargetPriors,
			final NetworkData networkData,
			Params params)
	{
		for (int conditionNum=0; conditionNum<params.numConditions; ++conditionNum) {
			try {
				BufferedReader reader = new BufferedReader(new FileReader(dremTargetsFile + "_cond_" + conditionNum + ".targets"));
				String line;
	
				for (;(line = reader.readLine()) != null;) {
					String [] splitLine = line.split("\t");
					int target = networkData.geneNameToId.get(splitLine[0]);
					// TODO Remove this
					if (!networkData.allTargets.getSet().contains(target)) {
						continue;
					}
					realTargets[conditionNum].add(networkData.allTargets.get(target));
					if (splitLine.length > 1) {
						realTargetPriors[conditionNum][networkData.allTargets.get(target)] = Double.parseDouble(splitLine[1]);
					}
					else {
						realTargetPriors[conditionNum][networkData.allTargets.get(target)] = 1.;
					}
				}
				reader.close();
			}
			catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

	public static void setupTargets(
			final NetworkData networkData,
			ArraySet realTargets,
			final int numRandTargets,
			ArraySet randTargets,
			Random r)
	{
		final ArrayMapInt allTargets = networkData.allTargets;
		final int numAllTargets = networkData.numAllTargets;
		
		ArrayList<Integer> randomTargetCandidates = new ArrayList<Integer>(numRandTargets);
		for (int i=0; i<numAllTargets; ++i) {
			if (!realTargets.contains(i)) {
				randomTargetCandidates.add(i);
			}
		}
		assert randomTargetCandidates.size() == numAllTargets-realTargets.size();
		
		Collections.shuffle(randomTargetCandidates, r);
		for (int i=0; i<numRandTargets; ++i) {
			randTargets.add(randomTargetCandidates.get(i));
		}
	}
	
	public static double computeObj(
			final boolean[] pathActive,
			final double[] pathCost)
	{
		double obj = 0.;
		for (int i=0; i<pathActive.length; ++i) {
			if (pathActive[i]) {
				obj += pathCost[i];
			}
		}
		return obj;
	}
	
	public static int computeActive(final boolean[] pathActive)
	{
		int numActive = 0;
		for (int i=0; i<pathActive.length; ++i) {
			if (pathActive[i]) {
				++numActive;
			}
		}
		return numActive;
	}
	
	public static void optimizeFlow(
			boolean[] fixedEdges,
			byte[] edgeDir,
			final TIntHashSet targets,
			final NetworkData networkData,
			final Params params)
	{
	}
	
	public static void optimizeTabu(
			int tabuSize,
			int tabuMaxItr,
			boolean[] fixedEdges,
			byte[] edgeDir,
			double[] flipCost,
			boolean[] pathActive,
			double [] pathCost,
			final TIntHashSet targets,
			final PathData pathData,
			final NetworkData networkData,
			final double convergenceRatio)
	{
		int[] tabuList = new int[tabuSize];
		int tabuListStart = 0;
		ArraySet tabuSet = new ArraySet(networkData.numEdges);

		boolean[] bestPathActive = new boolean[pathData.paths.length];
		double bestObj = 0.;

		int [][] paths = pathData.paths;
		long t1 = System.currentTimeMillis();
		int itr=0;
		assert fixedEdges.length == flipCost.length : fixedEdges.length + "\t" + flipCost.length;
		assert flipCost.length == networkData.numEdges : flipCost.length + "\t" + networkData.numEdges;

		double curObj = computeObj(pathActive, pathCost);
		for (int tabuItr=0;; ++itr) {
			int bestFlip = -1;
			for (int i=0; i<flipCost.length; ++i) {
				if (!fixedEdges[i] && (bestFlip == -1 || flipCost[i] > flipCost[bestFlip])) {
					if (bestFlip == -1 && !tabuSet.contains(i)) {
						bestFlip = i;
					}
					else if (bestFlip != -1) {
						if (flipCost[i] > flipCost[bestFlip] 
								&& (flipCost[i] > 0.0 || !tabuSet.contains(i))) 
						{
							bestFlip = i;
						}
					}
				}
			}
			
			if (bestFlip == -1) {
				pathData.pathActive = bestPathActive;
				break;
			}
			else if (flipCost[bestFlip] < 0.0) {
				if (tabuItr == tabuMaxItr)
					break;
				++tabuItr;
				if (curObj > bestObj) {
					System.out.println("best cur Obj: " + curObj);
					bestObj = curObj;
					for (int i=0; i<pathData.pathActive.length; ++i) {
						bestPathActive[i] = pathData.pathActive[i];
					}
				}
			}
			
			if (tabuSet.size() < tabuSize) {
				tabuList[tabuSet.size()] = bestFlip;
				tabuSet.add(bestFlip);
			}
			else {
				tabuSet.remove(tabuList[tabuListStart]);
				tabuSet.add(bestFlip);
				tabuList[tabuListStart] = bestFlip;
				++tabuListStart;
				
				if (tabuListStart == tabuSize) {
					tabuListStart = 0;
				}
			}
			
			curObj += flipCost[bestFlip];
			
			flip(bestFlip, 
					edgeDir, 
					pathActive, 
					flipCost,
					pathCost,
					targets,
					pathData,
					networkData);

			if (Global.XDEBUG) {
				double [] newFlipCost = new double[flipCost.length];
				computeAllFlipCosts(edgeDir, newFlipCost, pathActive, pathCost, pathData.paths, pathData.undirPaths);
				double diff = Math.abs(flipCost[bestFlip]-newFlipCost[bestFlip]);
				assert diff <= 0.0001 : diff;
				for (int i=0; i<flipCost.length; ++i) {
					if (fixedEdges[i])
						continue;
					diff = Math.abs(flipCost[i]-newFlipCost[i]);
					assert diff <= 0.0001 : diff + "\t" + i + "\t" + bestFlip;
				}
				for (int i=0; i<pathActive.length; ++i) {
					int tf = networkData.edgeToNodes[paths[i][paths[i].length-1]][1];
					if (!targets.contains(tf)) {
						continue;
					}
					boolean active = isPathActive(pathData.paths[i], edgeDir);
					assert active == pathActive[i] : active + "\t" + pathActive[i];
				}
			}
		}

		System.out.println("Done with itr=" + itr + " and obj " + computeObj(pathActive, pathCost) 
				+ " and numActive " + computeActive(pathActive)
				+ " in time " + (System.currentTimeMillis()-t1)/1000.0 + " s");
		
		if (Global.DEBUG) {
			computeAllFlipCosts(edgeDir, flipCost, pathActive, pathCost, pathData.paths, pathData.undirPaths);
			for (int i=0; i<flipCost.length; ++i) {
				if (fixedEdges[i])
					continue;
				assert flipCost[i] <= convergenceRatio : i + "\t" + flipCost[i];
			}
			for (int i=0; i<pathActive.length; ++i) {
				int tf = networkData.edgeToNodes[paths[i][paths[i].length-1]][1];
				if (!targets.contains(tf)) {
					continue;
				}
				boolean active = isPathActive(pathData.paths[i], edgeDir);
				assert active == pathActive[i] : active + "\t" + pathActive[i];
			}
			System.out.println("Reached Local Minimum");
		}
	}
	
	public static void optimize(
			boolean[] fixedEdges,
			byte[] edgeDir,
			double[] flipCost,
			boolean[] pathActive,
			double [] pathCost,
			final TIntHashSet targets,
			final PathData pathData,
			final NetworkData networkData,
			final double convergenceRatio)
	{
		int [][] paths = pathData.paths;
		long t1 = System.currentTimeMillis();
		int itr=0;
		assert fixedEdges.length == flipCost.length : fixedEdges.length + "\t" + flipCost.length;
		assert flipCost.length == networkData.numEdges : flipCost.length + "\t" + networkData.numEdges;
		for (;; ++itr) {
			int bestFlip = -1;
			for (int i=0; i<flipCost.length; ++i) {
				if (!fixedEdges[i] && (bestFlip == -1 || flipCost[i] > flipCost[bestFlip])) {
					bestFlip = i;
				}
			}
			
			if (bestFlip == -1 || flipCost[bestFlip] <= convergenceRatio) {
				if (bestFlip == -1) {
					System.out.println("converged totally");
				}
				else {
					System.out.println("converged " + flipCost[bestFlip] + " <= " + convergenceRatio);
				}
				break;
			}
			
			flip(bestFlip, 
					edgeDir, 
					pathActive, 
					flipCost,
					pathCost,
					targets,
					pathData,
					networkData);

			if (Global.XDEBUG) {
				double [] newFlipCost = new double[flipCost.length];
				computeAllFlipCosts(edgeDir, newFlipCost, pathActive, pathCost, pathData.paths, pathData.undirPaths);
				double diff = Math.abs(flipCost[bestFlip]-newFlipCost[bestFlip]);
				assert diff <= 0.0001 : diff;
				for (int i=0; i<flipCost.length; ++i) {
					if (fixedEdges[i])
						continue;
					diff = Math.abs(flipCost[i]-newFlipCost[i]);
					assert diff <= 0.0001 : diff + "\t" + i + "\t" + bestFlip;
				}
				for (int i=0; i<pathActive.length; ++i) {
					int tf = networkData.edgeToNodes[paths[i][paths[i].length-1]][1];
					if (!targets.contains(tf)) {
						continue;
					}
					boolean active = isPathActive(pathData.paths[i], edgeDir);
					assert active == pathActive[i] : active + "\t" + pathActive[i];
				}
			}
		}

		System.out.println("Done with itr=" + itr + " and obj " + computeObj(pathActive, pathCost) 
				+ " and numActive " + computeActive(pathActive)
				+ " in time " + (System.currentTimeMillis()-t1)/1000.0 + " s");
		
		if (Global.DEBUG) {
			computeAllFlipCosts(edgeDir, flipCost, pathActive, pathCost, pathData.paths, pathData.undirPaths);
			for (int i=0; i<flipCost.length; ++i) {
				if (fixedEdges[i])
					continue;
				assert flipCost[i] <= convergenceRatio : i + "\t" + flipCost[i];
			}
			for (int i=0; i<pathActive.length; ++i) {
				int tf = networkData.edgeToNodes[paths[i][paths[i].length-1]][1];
				if (!targets.contains(tf)) {
					continue;
				}
				boolean active = isPathActive(pathData.paths[i], edgeDir);
				assert active == pathActive[i] : active + "\t" + pathActive[i];
			}
			System.out.println("Reached Local Minimum");
		}
	}
	
	public final static class PathComp implements Comparator<Integer> {
		PathData pathData;
		int cond;

		public PathComp(PathData pathData, int cond) {
			this.pathData = pathData;
			this.cond = cond;
		}

		public int compare(Integer i1, Integer i2) {
			return Double.compare(pathData.pathCost[i2][cond], pathData.pathCost[i1][cond]);
		}
	}
	
	public final static class CompCost implements Comparator<Integer> {
		double [] cost;

		public CompCost(double [] cost) {
			this.cost = cost;
		}

		public int compare(Integer i1, Integer i2) {
			return Double.compare(cost[i2], cost[i1]);
		}
	}
	
	public static double setupPathCost(
			PathData pathData,
			NetworkData networkData,
			ArraySet [] realTargets,
			final double [][] realTargetPriors,
			ArraySet [] randTargets,
			Params params)
	{
		double maxPossible = 0.;
		int [][] paths = pathData.paths;
		double [][] pathCost = pathData.pathCost;
		double [] curIterPathCost = pathData.curIterPathCost;
		int numConditions = realTargetPriors.length;
		TIntHashSet [] sourceSet = networkData.sourceSet;

		int [][] class_count = new int[params.numClasses][networkData.numAllTargets];
		for (int tfIndex=0; tfIndex<networkData.numAllTargets; ++tfIndex) {
			for (int cond=0; cond<numConditions; ++cond) {
				if (realTargets[cond].contains(tfIndex)) {
					++class_count[params.cond_to_class[cond]][tfIndex];
				}
			}
			for (int cond=0; cond<numConditions; ++cond) {
				double factor = Math.pow(class_count[params.cond_to_class[cond]][tfIndex], params.tfShareFactor);
				System.out.println("class count for cond " + cond + ": " + networkData.geneIdToName[networkData.targetIndexToTarget[tfIndex]] + "\t" + class_count[params.cond_to_class[cond]][tfIndex] + "\t" + factor);
			}
		}

		for (int i=0; i<curIterPathCost.length; ++i) {
			int source = networkData.edgeToNodes[paths[i][0]][0];
			int tf = networkData.edgeToNodes[paths[i][paths[i].length-1]][1];
			curIterPathCost[i] = 0.;
			
			for (int cond=0; cond<numConditions; ++cond) {
				if (!sourceSet[cond].contains(source))
					continue;
				// convert to tfIndex because realTargets[cond] is an ArraySet
				int tfIndex = networkData.allTargets.get(tf);
				assert class_count[params.cond_to_class[cond]][tfIndex] <= params.numConditions : class_count[params.cond_to_class[cond]][tfIndex] + " for cond " + cond;
				double factor = Math.pow(class_count[params.cond_to_class[cond]][tfIndex], params.tfShareFactor);
				if (realTargets[cond].contains(tfIndex)) {
					assert class_count[params.cond_to_class[cond]][tfIndex] > 0;
					curIterPathCost[i] += factor*pathCost[i][cond]*realTargetPriors[cond][tfIndex];
				}
				else if (randTargets != null && randTargets[cond].contains(tfIndex)) {
					curIterPathCost[i] += pathCost[i][cond];
				}
			}
			maxPossible += curIterPathCost[i];
		}
		return maxPossible;
	}
	
	public static FlowData [] orientationMaxFlowMeasure(
		byte[] edgeDir,
		double[] flipCost,
		final NetworkData networkData,
		final Params params,
		ArraySet [] realTargets,
		final double[][] realTargetPriors,
		int[] numRandTargets,
		ArraySet[] randTargets,
		Random r
		)
	{
		FlowData [] flowData = null;
		final int numAllTargets = networkData.numAllTargets;
		final int numEdges = networkData.numEdges;
		final int numNodes = networkData.nodePriors.length;
		assert numNodes == networkData.edges.length : numNodes + " == " + networkData.nodePriors.length;

		// Count numTimes real targets appear
		int [][] class_count = new int[params.numClasses][networkData.numAllTargets];
		for (int cond=0; cond<params.numConditions; ++cond) {
			for (int tfIndex=0; tfIndex<networkData.numAllTargets; ++tfIndex) {
				if (realTargets[cond].contains(tfIndex)) {
					++class_count[params.cond_to_class[cond]][tfIndex];
				}
			}
		}

		try {
			IloCplex cplex = new IloCplex();
			cplex.setParam(IloCplex.IntParam.RootAlg, IloCplex.Algorithm.Barrier);

			TIntIntHashMap[] outEdges = new TIntIntHashMap[numNodes];
			TIntIntHashMap[] inEdges = new TIntIntHashMap[numNodes];
			TIntDoubleHashMap[] outEdgeCost = new TIntDoubleHashMap[numNodes];
			TIntDoubleHashMap[] inEdgeCost = new TIntDoubleHashMap[numNodes];
			TIntIntHashMap ppEdges = new TIntIntHashMap();
			int totalEdges = 0;
			for (int i=0; i<numEdges; ++i, ++totalEdges) {
				int source = networkData.edgeToNodes[i][0];
				int target = networkData.edgeToNodes[i][1];
				int edgeId = networkData.edgeToId[source].get(target);
				outEdges[source].put(target, totalEdges);
				inEdges[target].put(source, totalEdges);
				outEdgeCost[source].put(target, networkData.edgeCost[i]);
				inEdgeCost[target].put(source, networkData.edgeCost[i]);
				if (networkData.edgeType[edgeId] == (byte) 0) {
					ppEdges.put(totalEdges, edgeId);
					++totalEdges;
					outEdges[target].put(source, totalEdges);
					inEdges[source].put(target, totalEdges);
					outEdgeCost[target].put(source, networkData.edgeCost[i]);
					inEdgeCost[source].put(target, networkData.edgeCost[i]);
				}
			}

			IloIntVar[] edgeExistVars = cplex.intVarArray(totalEdges, 0, 1);
			IloNumVar[] edgeVars = cplex.numVarArray(totalEdges, 0., 1.);
			// Inflow for node == Outflow for node
			for (int curNode=0; curNode<numNodes; ++curNode) {
				IloNumExpr inflow = cplex.constant(0.);
				IloNumExpr outflow = cplex.constant(0.);

				for (TIntIntIterator inNodeIt=inEdges[curNode].iterator(); inNodeIt.hasNext(); inNodeIt.advance()) {
					int inNode = inNodeIt.key();
					inflow = cplex.sum(edgeVars[inNodeIt.value()], inflow);
				}
				for (TIntIntIterator outNodeIt=outEdges[curNode].iterator(); outNodeIt.hasNext(); outNodeIt.advance()) {
					int outNode = outNodeIt.key();
					outflow = cplex.sum(edgeVars[outNodeIt.value()], outflow);
				}

				cplex.addEq(inflow, outflow);
			}
			
			for (TIntIntIterator it=ppEdges.iterator(); it.hasNext(); it.advance()) {
				int edge = it.key();
				cplex.addLe(edgeVars[edge], edgeExistVars[edge]);
				cplex.addLe(edgeVars[edge+1], edgeExistVars[edge+1]);
				cplex.addEq(cplex.sum(edgeExistVars[edge], edgeExistVars[edge+1]), 0);
			}
			
			TIntHashSet curSources = new TIntHashSet(); 
			TIntHashSet curTargets = new TIntHashSet();
			IloNumVar[] auxTargetVars = cplex.numVarArray(networkData.numAllTargets, 0., 0.);
			double [] targetWeights = new double[networkData.numAllTargets];
			double avgTargetWeight = 0.;
			int numCurTargets = 0;

			for (int cond=0; cond<params.numConditions; ++cond) {
				randTargets[cond] = new ArraySet(numAllTargets);
				// Sets up the random curTargets
				if (numRandTargets[cond] > numAllTargets-realTargets[cond].size()) {
					throw new IllegalStateException(numRandTargets[cond] + " > " + numAllTargets
							+ " - " + realTargets[cond].size());
				}
	
				// setup random curTargets
				setupTargets(
						networkData,
						realTargets[cond],
						numRandTargets[cond],
						randTargets[cond],
						r);
	
				for (int i=0; i<randTargets[cond].size(); ++i) {
					targetWeights[randTargets[cond].get(i)] += 1.0;
				}
				for (int i=0; i<realTargets[cond].size(); ++i) {
					int tfIndex = realTargets[cond].get(i);
					double factor = Math.pow(params.tfShareFactor, class_count[params.cond_to_class[cond]][tfIndex]);
					targetWeights[tfIndex] += factor*realTargetPriors[cond][tfIndex];
				}
				
				curSources.addAll(networkData.sourceSet[cond]);
				assert realTargets[cond].size() == networkData.numAllTargets : realTargets[cond].size() + "==" + networkData.numAllTargets;
				for (int tfIndex=0; tfIndex<networkData.numAllTargets; ++tfIndex) {
					auxTargetVars[tfIndex].setUB(0.0);
					if (realTargets[cond].contains(tfIndex)) {
						avgTargetWeight += targetWeights[tfIndex];
						curTargets.add(tfIndex);
					}
					else if (randTargets[cond].contains(tfIndex)) {
						avgTargetWeight += 1.0;
						curTargets.add(tfIndex);
					}
				}
			}

			avgTargetWeight /= curTargets.size();
			IloNumVar[] auxSourceVars = cplex.numVarArray(curSources.size(), 0., avgTargetWeight);
			for (int tfIndex=0; tfIndex<networkData.numAllTargets; ++tfIndex) {
				if (curTargets.contains(tfIndex)) {
					auxTargetVars[tfIndex].setUB(targetWeights[tfIndex]);
				}
			}

			IloNumExpr auxSourceFlow = cplex.constant(0.);
			IloNumExpr auxTargetFlow = cplex.constant(0.);
			for (int i=0; i<auxSourceVars.length; ++i) {
				auxSourceFlow = cplex.sum(auxSourceVars[i], auxSourceFlow);
			}
			for (int i=0; i<auxTargetVars.length; ++i) {
				auxTargetFlow = cplex.sum(auxTargetVars[i], auxTargetFlow);
			}
			cplex.addEq(auxSourceFlow, auxTargetFlow);

			
			// Cost of flow
			IloNumExpr edgeCost = cplex.constant(0.0);
			for (int i=0; i<numEdges; ++i) {
				edgeCost = cplex.sum(-Math.log(networkData.edgeCost[i]), edgeCost);
			}
			IloNumExpr sourcePenalty = cplex.constant(0.);
			for (int i=0; i<auxSourceVars.length; ++i) {
				sourcePenalty  = cplex.sum(auxSourceVars[i], sourcePenalty);
			}
			sourcePenalty = cplex.prod(params.maxFlowSourcePenalty, sourcePenalty);

			IloNumExpr obj = cplex.diff(edgeCost, sourcePenalty);

			cplex.addMinimize(obj);

			boolean solve = cplex.solve();
			assert solve;
			
			double[] edgeExists = cplex.getValues(edgeExistVars);
			for (TIntIntIterator it=ppEdges.iterator(); it.hasNext(); it.advance()) {
				int edge = it.key();
				assert Math.abs(edgeExists[edge]+edgeExists[edge+1]-1.0) < Global.tol;
				if (Math.abs(edgeExists[edge]-1.0) < Global.tol) {
					edgeDir[it.value()] = 1;
				}
				else {
					edgeDir[it.value()] = 0;
				}
			}
			
			for (int cond=0; cond<params.numConditions; ++cond) {
			}

			flowData = maxFlowMeasure(
					edgeDir, 
					flipCost, 
					networkData, 
					params, 
					realTargets,
					realTargetPriors,
					numRandTargets,
					randTargets,
					r);
		}
		catch (IloException e) {
			System.err.println("Concert exception " + e);
		}
		return flowData;
	}
	
	public static FlowData [] maxFlowMeasure(
			byte[] edgeDir,
			double[] flipCost,
			final NetworkData networkData,
			final Params params,
			ArraySet [] realTargets,
			final double[][] realTargetPriors,
			int[] numRandTargets,
			ArraySet[] randTargets,
			Random r
			)
	{
		FlowData [] flowData = new FlowData[params.numConditions];
		final int numAllTargets = networkData.numAllTargets;
		final int numEdges = networkData.numEdges;
		final int numNodes = networkData.nodePriors.length;
		assert numNodes == networkData.edges.length : numNodes + " == " + networkData.nodePriors.length;

		// Count numTimes real targets appear
		int [][] class_count = new int[params.numClasses][networkData.numAllTargets];
		for (int cond=0; cond<params.numConditions; ++cond) {
			for (int tfIndex=0; tfIndex<networkData.numAllTargets; ++tfIndex) {
				if (realTargets[cond].contains(tfIndex)) {
					++class_count[params.cond_to_class[cond]][tfIndex];
				}
			}
		}

		try {
			IloCplex cplex = new IloCplex();
			cplex.setParam(IloCplex.IntParam.RootAlg, IloCplex.Algorithm.Barrier);

			TIntIntHashMap[] outEdges = new TIntIntHashMap[numNodes];
			TIntIntHashMap[] inEdges = new TIntIntHashMap[numNodes];
			TIntDoubleHashMap[] outEdgeCost = new TIntDoubleHashMap[numNodes];
			TIntDoubleHashMap[] inEdgeCost = new TIntDoubleHashMap[numNodes];
			int totalEdges = 0;
			if (edgeDir == null) {
				for (int i=0; i<numEdges; ++i, ++totalEdges) {
					int source = networkData.edgeToNodes[i][0];
					int target = networkData.edgeToNodes[i][1];
					outEdges[source].put(target, totalEdges);
					inEdges[target].put(source, totalEdges);
					outEdgeCost[source].put(target, networkData.edgeCost[i]);
					inEdgeCost[target].put(source, networkData.edgeCost[i]);
					if (networkData.edgeType[networkData.edgeToId[source].get(target)] == (byte) 0) {
						++totalEdges;
						outEdges[target].put(source, totalEdges);
						inEdges[source].put(target, totalEdges);
						outEdgeCost[target].put(source, networkData.edgeCost[i]);
						inEdgeCost[source].put(target, networkData.edgeCost[i]);
					}
				}
			}
			else {
				assert edgeDir.length == numEdges : edgeDir.length + " == " + numEdges;
				for (int i=0; i<numEdges; ++i, ++totalEdges) {
					int source = networkData.edgeToNodes[i][0];
					int target = networkData.edgeToNodes[i][1];
					int edgeId = networkData.edgeToId[source].get(target);
					
					if (edgeDir[edgeId] == 1) {
						outEdges[source].put(target, totalEdges);
						inEdges[target].put(source, totalEdges);
						outEdgeCost[source].put(target, networkData.edgeCost[i]);
						inEdgeCost[target].put(source, networkData.edgeCost[i]);
					}
					else {
						outEdges[target].put(source, totalEdges);
						inEdges[source].put(target, totalEdges);
						outEdgeCost[target].put(source, networkData.edgeCost[i]);
						inEdgeCost[source].put(target, networkData.edgeCost[i]);
					}
				}
			}

			IloNumVar[] edgeVars = cplex.numVarArray(totalEdges, 0., 1.);
			// Inflow for node == Outflow for node
			for (int curNode=0; curNode<numNodes; ++curNode) {
				IloNumExpr inflow = cplex.constant(0.);
				IloNumExpr outflow = cplex.constant(0.);

				for (TIntIntIterator inNodeIt=inEdges[curNode].iterator(); inNodeIt.hasNext(); inNodeIt.advance()) {
					int inNode = inNodeIt.key();
					inflow = cplex.sum(edgeVars[inNodeIt.value()], inflow);
				}
				for (TIntIntIterator outNodeIt=outEdges[curNode].iterator(); outNodeIt.hasNext(); outNodeIt.advance()) {
					int outNode = outNodeIt.key();
					outflow = cplex.sum(edgeVars[outNodeIt.value()], outflow);
				}

				cplex.addEq(inflow, outflow);
			}

			for (int cond=0; cond<params.numConditions; ++cond) {
				LinkedList<IloAddable> addedVars = new LinkedList<IloAddable>();
				LinkedList<IloAddable> addedConstraints = new LinkedList<IloAddable>();
				randTargets[cond] = new ArraySet(numAllTargets);
				// Sets up the random curTargets
				if (numRandTargets[cond] > numAllTargets-realTargets[cond].size()) {
					throw new IllegalStateException(numRandTargets[cond] + " > " + numAllTargets
							+ " - " + realTargets[cond].size());
				}
	
				// setup random curTargets
				setupTargets(
						networkData,
						realTargets[cond],
						numRandTargets[cond],
						randTargets[cond],
						r);
	
				double [] targetWeights = new double[networkData.numAllTargets];
				for (int i=0; i<randTargets[cond].size(); ++i) {
					targetWeights[randTargets[cond].get(i)] += 1.0;
				}
				for (int i=0; i<realTargets[cond].size(); ++i) {
					int tfIndex = realTargets[cond].get(i);
					double factor = Math.pow(params.tfShareFactor, class_count[params.cond_to_class[cond]][tfIndex]);
					targetWeights[tfIndex] += factor*realTargetPriors[cond][tfIndex];
				}

				// XXX

				TIntHashSet curSources = networkData.sourceSet[cond];
				IloNumVar[] auxTargetVars = cplex.numVarArray(networkData.numAllTargets, 0., 0.);
				
				assert realTargets[cond].size() == networkData.numAllTargets : realTargets[cond].size() + "==" + networkData.numAllTargets;
				double avgTargetWeight = 0.;
				int numCurTargets = 0;
				for (int tfIndex=0; tfIndex<networkData.numAllTargets; ++tfIndex) {
					addedVars.add(auxTargetVars[tfIndex]);
					if (realTargets[cond].contains(tfIndex)) {
						auxTargetVars[tfIndex].setUB(targetWeights[tfIndex]);
						avgTargetWeight += targetWeights[tfIndex];
						++numCurTargets;
					}
					else if (randTargets[cond].contains(tfIndex)) {
						auxTargetVars[tfIndex].setUB(1.0);
						avgTargetWeight += 1.0;
						++numCurTargets;
					}
					else {
						auxTargetVars[tfIndex].setUB(0.0);
					}
				}
				avgTargetWeight /= numCurTargets;
				IloNumVar[] auxSourceVars = cplex.numVarArray(curSources.size(), 0., avgTargetWeight);

				IloNumExpr auxSourceFlow = cplex.constant(0.);
				IloNumExpr auxTargetFlow = cplex.constant(0.);
				for (int i=0; i<auxSourceVars.length; ++i) {
					addedVars.add(auxSourceVars[i]);
					auxSourceFlow = cplex.sum(auxSourceVars[i], auxSourceFlow);
				}
				for (int i=0; i<auxTargetVars.length; ++i) {
					auxTargetFlow = cplex.sum(auxTargetVars[i], auxTargetFlow);
				}
				addedConstraints.add(cplex.addEq(auxSourceFlow, auxTargetFlow));
				
				for (int tfIndex=0; tfIndex<numAllTargets; ++tfIndex) {
					IloNumExpr targetInFlow = cplex.constant(0.0);
					
					for(TIntIntIterator it=inEdges[networkData.targetIndexToTarget[tfIndex]].iterator();
							it.hasNext();
							it.advance())
					{
						targetInFlow = cplex.sum(edgeVars[it.value()], targetInFlow);
					}
					
					addedConstraints.add(cplex.addEq(targetInFlow, auxTargetVars[tfIndex]));
				}
				
				int curSource=0;
				for (TIntIterator it=curSources.iterator(); it.hasNext(); ++curSource) {
					int source = it.next();
					IloNumExpr sourceOutFlow = cplex.constant(0.0);
					for(TIntIntIterator it2=outEdges[source].iterator(); it2.hasNext(); it2.advance())
					{
						sourceOutFlow = cplex.sum(edgeVars[it2.value()], sourceOutFlow);
					}
					addedConstraints.add(cplex.addEq(sourceOutFlow, auxSourceVars[curSource]));
				}

				// Cost of flow -- Objective
				IloNumExpr edgeCost = cplex.constant(0.);
				for (int i=0; i<numEdges; ++i) {
					edgeCost = cplex.sum(-Math.log(networkData.edgeCost[i]), edgeCost);
				}
				IloNumExpr sourcePenalty = cplex.constant(0.);
				for (int i=0; i<auxSourceVars.length; ++i) {
					sourcePenalty  = cplex.sum(auxSourceVars[i], sourcePenalty);
				}
				sourcePenalty = cplex.prod(params.maxFlowSourcePenalty, sourcePenalty);

				IloNumExpr obj = cplex.diff(edgeCost, sourcePenalty);

				addedConstraints.add(cplex.addMinimize(obj));

				boolean solve = cplex.solve();
				assert solve;

				flowData[cond] = new FlowData();
				flowData[cond].edgeFlow = cplex.getValues(edgeVars);
				flowData[cond].targetFlow = cplex.getValues(auxTargetVars);
				flowData[cond].nodeFlow = new double[numNodes];
				
				for (int i=0; i<numNodes; ++i) {
					for (TIntIntIterator it=inEdges[i].iterator(); it.hasNext(); it.advance()) {
						flowData[cond].nodeFlow[i] += flowData[cond].edgeFlow[it.value()];
					}
				}
				
				for (IloAddable v : addedVars) {
					cplex.remove(v);
				}
				for (IloAddable c : addedConstraints) {
					cplex.remove(c);
				}
			}
		}
		catch (IloException e) {
			System.err.println("Concert exception " + e);
		}
		return flowData;
	}
	
	public static PathData runOrientation(
			byte[] edgeDir,
			double[] flipCost,
			final NetworkData networkData,
			final Params params,
			ArraySet [] realTargets,
			final double[][] realTargetPriors,
			int[] numRandTargets,
			ArraySet[] randTargets,
			Random r
			)
	{
		TIntHashSet targets = new TIntHashSet();
		int numAllTargets = networkData.numAllTargets;

		for (int cond=0; cond<params.numConditions; ++cond) {
			randTargets[cond] = new ArraySet(numAllTargets);
			// Sets up the random targets
			if (numRandTargets[cond] > numAllTargets-realTargets[cond].size()) {
				throw new IllegalStateException(numRandTargets[cond] + " > " + numAllTargets
						+ " - " + realTargets[cond].size());
			}

			// setup random targets
			setupTargets(
					networkData,
					realTargets[cond],
					numRandTargets[cond],
					randTargets[cond],
					r);

			for (int j=0; j<realTargets[cond].size(); ++j) {
				targets.add(networkData.targetIndexToTarget[realTargets[cond].get(j)]);
			}
			for (int j=0; j<randTargets[cond].size(); ++j) {
				targets.add(networkData.targetIndexToTarget[randTargets[cond].get(j)]);
			}
		}
		
		double [] targetWeights = new double[networkData.numAllTargets];
		for (int cond=0; cond<params.numConditions; ++cond) {
			for (int i=0; i<randTargets[cond].size(); ++i) {
				targetWeights[randTargets[cond].get(i)] += 1.0;
			}
			for (int i=0; i<realTargets[cond].size(); ++i) {
				assert realTargetPriors[cond][realTargets[cond].get(i)] >= 1-0.00001;
				int tfIndex = realTargets[cond].get(i);
				//int tf = networkData.targetIndexToTarget[tfIndex];
				targetWeights[realTargets[cond].get(i)] += realTargetPriors[cond][realTargets[cond].get(i)];
			}
		}

		PathData pathData = new PathData();
		try {
			readTargetsAndPaths(targets, targetWeights, params.pathEnumBound, pathData, networkData.nodePriors.length, networkData, params);
		}
		catch (Exception e) {
			e.printStackTrace();
		}
			
		// Setup init path costs
		int [][] paths = pathData.paths;
		pathData.pathCost = new double[paths.length][params.numConditions];
		for (int i=0; i<paths.length; ++i) {
			double cost = 1.;
			int source = networkData.edgeToNodes[paths[i][0]][0];
			for (int j=0; j<paths[i].length; ++j) {
				cost *= networkData.edgeCost[paths[i][j]];
			}
			for (int cond=0; cond<params.numConditions; ++cond) {
				if (networkData.sourceSet[cond].contains(source)) {
					double condCost = cost*networkData.nodePriors[source];
					for (int k=0; k<paths[i].length; ++k) {
						int pathNode = networkData.edgeToNodes[paths[i][k]][1];
						condCost *= networkData.nodePriors[pathNode];
					}
					pathData.pathCost[i][cond] = condCost;
				}
				else {
					pathData.pathCost[i][cond] = 0;
				}
			}
		}

		pathData.pathActive = new boolean[paths.length];
		// Sets up the path costs for the current iteration
		double maxPossible = setupPathCost(pathData, networkData, realTargets, realTargetPriors, randTargets, params);
		System.out.println("Max Possible Obj: " + maxPossible);

		// Sets up init edge dirs and the init flip costs
		boolean [] fixedEdges = randomFlipping(edgeDir, flipCost, pathData.pathActive, pathData.curIterPathCost, pathData, networkData, targets, r);

		// Local search
		optimize(fixedEdges, edgeDir, flipCost, pathData.pathActive, pathData.curIterPathCost, targets, pathData, networkData, params.convergenceRatio);
		//optimizeTabu(params.tabuSize, params.tabuMaxItr, fixedEdges, edgeDir, flipCost, pathData.pathActive, pathData.curIterPathCost, targets, pathData, networkData, params.convergenceRatio);
		
		return pathData;
	}

	public static void computeTfPriorsMaxFlow(
			byte[] edgeDir,
			double[] flipCost,
			final NetworkData networkData,
			final Params params,
			final int sdremIteration,
			final String rankingMetric,
			Random r,
			ArraySet [] realTargets,
			double [][] realTargetPriors,
			double [][] realTargetPercentile)
	{
		final ArrayMapInt allTargets = networkData.allTargets;
		final int numAllTargets = networkData.numAllTargets;

		int [] numRandTargets = new int[params.numConditions];
		ArraySet [] randTargets  = new ArraySet[params.numConditions];
		
		for (int i=0; i<params.numConditions; ++i) {
			numRandTargets[i] = (int) ((double) realTargets[i].size() * params.randTargThreshold);
		}
		assert numRandTargets[0] > 0;

		double [][] allScores = new double[params.numConditions]
				[params.orientationRestarts*numAllTargets];
		for (int i=0; i<params.numConditions; ++i) {
			allScores[i] = new double[params.orientationRestarts*numAllTargets];
		}
		
		double [][] realAvgScores = new double[params.numConditions][numAllTargets];
		
		for (int iter=0; iter<params.orientationRestarts; ++iter) {
			FlowData [] flowData = null;
			if (params.networkMethod.equals("maxflow.orientation")) {
				assert edgeDir != null;
				flowData = orientationMaxFlowMeasure(
						edgeDir, 
						flipCost, 
						networkData, 
						params, 
						realTargets,
						realTargetPriors,
						numRandTargets,
						randTargets,
						r);
			}
			else {
				flowData = maxFlowMeasure(
						null, 
						flipCost, 
						networkData, 
						params, 
						realTargets,
						realTargetPriors,
						numRandTargets,
						randTargets,
						r);
			}
			
			for (int cond=0; cond<params.numConditions; ++cond) {
				for (int tfIndex=0; tfIndex<networkData.numAllTargets; ++tfIndex) {
					if (realTargets[cond].contains(tfIndex)) {
						realAvgScores[cond][tfIndex] += flowData[cond].targetFlow[tfIndex];
					}
					else if (randTargets[cond].contains(tfIndex)) {
						allScores[cond][tfIndex+numAllTargets*iter] += 
								flowData[cond].targetFlow[tfIndex];
					}
				}
			}
		}
		System.out.println("Done orientation restarts");
		
		processTfFlowScores(
				sdremIteration,
				rankingMetric,
				realTargets,
				randTargets,
				numRandTargets,
				realAvgScores,
				allScores,
				networkData,
				params);
		
	}
	
	public static void processTfFlowScores(
			int sdremIteration,
			String rankingMetric,
			ArraySet[] realTargets,
			ArraySet[] randTargets,
			int[] numRandTargets,
			double[][] realAvgScores,
			double[][] allScores,
			NetworkData networkData,
			Params params)
	{
		int numAllTargets = networkData.numAllTargets;
		int [][] realTargetRanks = new int[params.numConditions][numAllTargets];
		double [][] realTargetPercentile = new double[params.numConditions][numAllTargets];
		
		// Average real scores and sort all random scores for each condition
		for (int cond=0; cond<params.numConditions; ++cond) {
			Arrays.sort(allScores[cond]);
			for (int j=0; j<numAllTargets; ++j) {
				realAvgScores[cond][j] /= params.orientationRestarts;
			}
		}
		
		// Compute real target ranks
		for (int cond=0; cond<params.numConditions; ++cond) {
			System.out.println("allScores: " + allScores[cond][0] + "\t" + allScores[cond][allScores[cond].length-1]);
			int totalTargets = numRandTargets[cond]*params.orientationRestarts;
			for (int j=0; j<realTargets[cond].size(); ++j) {
				int tfIndex = realTargets[cond].get(j);
				int k=allScores[cond].length-1;
				for (; k>=0; --k) {
					if (realAvgScores[cond][tfIndex] > allScores[cond][k] )
						break;
				}
				realTargetRanks[cond][tfIndex] = k;
				realTargetPercentile[cond][tfIndex] = ((double) k-(allScores[cond].length-totalTargets)+1)/((double) totalTargets);
				//System.out.println(networkData.geneIdToName[networkData.targetIndexToTarget[tfIndex]] + "\t" + realTargetPercentile[cond][tfIndex] + "\t" + k + "\t" + realAvgScores[cond][tfIndex]);
			}
		}
		
		System.out.println("computed ranks");
		
		// Write scores_itr files
		if (!params.modelDir.equals("")) {
			for (int i=0; i<params.numConditions; ++i) {
				try {
					String out = params.modelDir + "scores_itr" + sdremIteration + "_r" + params.randTargThreshold + 
					"_" + rankingMetric + "_" 
						+ params.orientationRestarts + "_cond_" + i + ".txt";
					PrintWriter writer = new PrintWriter(new FileWriter(out));
					writer.println("Target\tAverage score\t# Rand greater than\t% Rand greater than");
		
					for (int j=0; j<realTargets[i].size(); ++j) {
						int tfIndex = realTargets[i].get(j);
						int tf = networkData.targetIndexToTarget[tfIndex];
						writer.println(networkData.geneIdToName[tf] + "\t" + realTargetRanks[i][tfIndex] +
								"\t" + realTargetPercentile[i][tfIndex]);
					}
				}
				catch (Exception e) {
					e.printStackTrace();
				}
			}
		}
	}
	
	public static void computeTfPriorsPathEnum(
			byte[] edgeDir,
			double[] flipCost,
			final NetworkData networkData,
			final Params params,
			final int sdremIteration,
			final String rankingMetric,
			Random r,
			ArraySet [] realTargets,
			double [][] realTargetPriors,
			double [][] realTargetPercentile)
	{
		final ArrayMapInt allTargets = networkData.allTargets;
		final int numAllTargets = networkData.numAllTargets;

		int [] pathThreshold = new int[params.numConditions];
		int [] numRandTargets = new int[params.numConditions];
		ArraySet [] randTargets  = new ArraySet[params.numConditions];
		
		for (int i=0; i<params.numConditions; ++i) {
			numRandTargets[i] = (int) ((double) realTargets[i].size() * params.randTargThreshold);
			assert numRandTargets[i] > 0;
			pathThreshold[i] = params.topRankedPaths;
			if (pathThreshold[i] < 0) {
				pathThreshold[i] = 5 * (realTargets[i].size() + numRandTargets[i]);
				System.out.println("Automatically setting path threshold for condition " + i);
			}
		}
		assert numRandTargets[0] > 0;

		double [][] allScores = new double[params.numConditions]
				[params.orientationRestarts*numAllTargets];
		for (int i=0; i<params.numConditions; ++i) {
			allScores[i] = new double[params.orientationRestarts*numAllTargets];
		}
		
		for (int i=0; i<params.numConditions; ++i) {
			System.out.println("Number of top paths used in calculations in condition " + i + ": " + pathThreshold[i]);
		}
		double [][] realAvgScores = new double[params.numConditions][numAllTargets];
		
		for (int iter=0; iter<params.orientationRestarts; ++iter) {
			PathData pathData = runOrientation(
					edgeDir, 
					flipCost, 
					networkData, 
					params, 
					realTargets,
					realTargetPriors,
					numRandTargets,
					randTargets,
					r);
			
			int[][] paths = pathData.paths;
			boolean[] pathActive = pathData.pathActive;
			System.out.println("Done restart " + iter + " orientation");
			
			// Total real scores over all orientation restarts and store the random
			// scores for each restart
			Integer [] sortedPaths = new Integer[pathData.paths.length];
			for (int i=0; i<sortedPaths.length; ++i) {
				sortedPaths[i] = i;
			}
			for (int j=0; j<params.numConditions; ++j) {
				Arrays.sort(sortedPaths, new PathComp(pathData, j));
				int thresh = Math.min(pathThreshold[j], sortedPaths.length);
				for (int i=0, k=0; i<thresh && k<sortedPaths.length; ++k) {
					int pathId = sortedPaths[k];
					int source = networkData.edgeToNodes[paths[pathId][0]][0];
					if (!pathActive[pathId] || !networkData.sourceSet[j].contains(source)) {
						continue;
					}
					++i;
					int tf = networkData.edgeToNodes[paths[pathId][paths[pathId].length-1]][1];
					int tfIndex = allTargets.get(tf);
					if (realTargets[j].contains(tfIndex)) {
						realAvgScores[j][tfIndex] += pathData.pathCost[pathId][j];
					}
					else if (randTargets[j].contains(tfIndex)) {
						allScores[j][tfIndex+numAllTargets*iter] += 
								pathData.pathCost[pathId][j];
					}
				}

			}
		}
		System.out.println("Done all restarts");
		// Done orientation restarts

		int [][] realTargetRanks = new int[params.numConditions][numAllTargets];
		
		// Average real scores and sort all random scores for each condition
		for (int cond=0; cond<params.numConditions; ++cond) {
			Arrays.sort(allScores[cond]);
			for (int j=0; j<numAllTargets; ++j) {
				realAvgScores[cond][j] /= params.orientationRestarts;
			}
		}
		
		// Compute real target ranks
		for (int cond=0; cond<params.numConditions; ++cond) {
			System.out.println("allScores: " + allScores[cond][0] + "\t" + allScores[cond][allScores[cond].length-1]);
			int totalTargets = numRandTargets[cond]*params.orientationRestarts;
			for (int j=0; j<realTargets[cond].size(); ++j) {
				int tfIndex = realTargets[cond].get(j);
				int k=allScores[cond].length-1;
				for (; k>=0; --k) {
					if (realAvgScores[cond][tfIndex] > allScores[cond][k] )
						break;
				}
				realTargetRanks[cond][tfIndex] = k;
				realTargetPercentile[cond][tfIndex] = ((double) k-(allScores[cond].length-totalTargets)+1)/((double) totalTargets);
				//System.out.println(networkData.geneIdToName[networkData.targetIndexToTarget[tfIndex]] + "\t" + realTargetPercentile[cond][tfIndex] + "\t" + k + "\t" + realAvgScores[cond][tfIndex]);
			}
		}
		
		System.out.println("computed ranks");
		
		// Write scores_itr files
		if (!params.modelDir.equals("")) {
			for (int i=0; i<params.numConditions; ++i) {
				try {
					String out = params.modelDir + "scores_maxflow_itr" + sdremIteration + "_r" + params.randTargThreshold + 
					"_" + rankingMetric + "_" 
						+ params.orientationRestarts + "_" + pathThreshold[i] + "_cond_" + i + ".txt";
					PrintWriter writer = new PrintWriter(new FileWriter(out));
					writer.println("Target\tAverage score\t# Rand greater than\t% Rand greater than");
		
					for (int j=0; j<realTargets[i].size(); ++j) {
						int tfIndex = realTargets[i].get(j);
						int tf = networkData.targetIndexToTarget[tfIndex];
						writer.println(networkData.geneIdToName[tf] + "\t" + realTargetRanks[i][tfIndex] +
								"\t" + realTargetPercentile[i][tfIndex]);
					}
				}
				catch (Exception e) {
					e.printStackTrace();
				}
			}
		}
	}

	public static void rankedPathScoreMaxFlow(
			byte[] edgeDir,
			double[] flipCost,
			double[][] realTargetPriors,
			int sdremIteration,
			String rankingMetric,
			ArraySet [] realTargets,
			double[] nodePathPercent,
			NetworkData networkData,
			Params params,
			Random r
			)
	{
		int numNodes = networkData.numNodes;
		assert numNodes == nodePathPercent.length : numNodes + "\t" + nodePathPercent.length;
		int [] numRandTargets = new int[params.numConditions];
		ArraySet [] randTargets  = new ArraySet[params.numConditions];
		ArrayMapInt allTargets = networkData.allTargets;

		FlowData [] flowData = null;
		if (params.networkMethod.equals("maxflow.orientation")) {
			assert edgeDir != null;
			flowData = orientationMaxFlowMeasure(
					edgeDir, 
					flipCost, 
					networkData, 
					params, 
					realTargets,
					realTargetPriors,
					numRandTargets,
					randTargets,
					r);
		}
		else {
			flowData = maxFlowMeasure(
					null, 
					flipCost, 
					networkData, 
					params, 
					realTargets,
					realTargetPriors,
					numRandTargets,
					randTargets,
					r);
		}
		System.out.println("Done real orientation");


		for (int cond=0; cond<params.numConditions; ++cond) {
			int pathThreshold = params.topRankedPaths;
			if (pathThreshold < 0) {
				pathThreshold = 5 * realTargets[cond].size();
				System.out.println("Automatically setting path threshold for condition " + cond);
			}


			double [] allPathsScore = flowData[cond].nodeFlow;
			double totalAllPathScore = 0.;
			
			for (int i=0; i<numNodes; ++i) {
				totalAllPathScore += allPathsScore[i];
			}

			if(!params.modelDir.equals("")) {
				try {
					// Write node scores
					String out = params.modelDir + "nodeScores_maxflow_itr" + sdremIteration + 
							"_" + rankingMetric + "_" 
							+ params.orientationRestarts + "_cond_" + cond+".txt";
					PrintWriter writer = new PrintWriter(new FileWriter(out));
					writer.println("ORF name\tStandard name\tSource\tTarget\t" + 
							"Weighted % top " + pathThreshold + " paths through node\t" +
							"Weighted % all paths through node");
	
					for (int j=0; j<numNodes; ++j) {
						assert !networkData.geneIdToName[j].trim().equals("");
						StringBuffer buf = new StringBuffer(
								networkData.geneIdToName[j] + "\t" + 
										networkData.geneIdToName[j]);
						if (networkData.sourceSet[cond].contains(j)) {
							buf.append("\t Y");
						}
						else {
							buf.append("\t N");
						}
						if (allTargets.getSet().contains(j) && realTargets[cond].contains(allTargets.get(j))) {
							buf.append("\t Y");
						}
						else {
							buf.append("\t N");
						}
						buf.append("\t" + allPathsScore[j]);
						writer.println(buf.toString());
					}
					writer.close();
				}
				catch (Exception e) {
					e.printStackTrace();
				}
			}
		}
	}
	
	public static void rankedPathScorePathEnum(
			byte[] edgeDir,
			double[] flipCost,
			double[][] realTargetPriors,
			int sdremIteration,
			String rankingMetric,
			ArraySet [] realTargets,
			double[] nodePathPercent,
			NetworkData networkData,
			Params params,
			Random r
			)
	{
		int numNodes = networkData.geneNameToId.size();
		assert numNodes == nodePathPercent.length : numNodes + "\t" + nodePathPercent.length;
		int [] numRandTargets = new int[params.numConditions];
		ArraySet [] randTargets  = new ArraySet[params.numConditions];
		ArrayMapInt allTargets = networkData.allTargets;

		PathData pathData = runOrientation(
				edgeDir, 
				flipCost, 
				networkData, 
				params, 
				realTargets,
				realTargetPriors,
				numRandTargets,
				randTargets,
				r);
		System.out.println("Done real orientation");

		int [][] paths = pathData.paths;
		double [] curIterPathCost = pathData.curIterPathCost;

		Integer [] sortedPaths = new Integer[pathData.paths.length];
		for (int i=0; i<sortedPaths.length; ++i) {
			sortedPaths[i] = i;
		}

		try {
			String out = params.modelDir + "edgeDir_itr" + sdremIteration + ".txt";

			PrintWriter writer = new PrintWriter(new FileWriter(out));
			
			int[][] edges = networkData.edges;
			assert networkData.numEdges > 0;
			String[] geneIdToName = networkData.geneIdToName;
			for (int source=0; source<edges.length; ++source) {
				for (int targetIndex=0; targetIndex<edges[source].length; ++targetIndex) {
					int target = edges[source][targetIndex];
					writer.write(geneIdToName[source] + "\t"
							+ geneIdToName[target] + "\t"
							+ edgeDir[networkData.edgeToId[source].get(target)] + "\n");
				}
			}
			writer.close();
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		
					
		
		for (int cond=0; cond<params.numConditions; ++cond) {
			int pathThreshold = params.topRankedPaths;
			if (pathThreshold < 0) {
				pathThreshold = 5 * realTargets[cond].size();
				System.out.println("Automatically setting path threshold for condition " + cond);
			}

			for (int i=0; i<curIterPathCost.length; ++i) {
				int source = networkData.edgeToNodes[paths[i][0]][0];
				int tf = networkData.edgeToNodes[paths[i][paths[i].length-1]][1];
				if (networkData.sourceSet[cond].contains(source)) {
					curIterPathCost[i] = pathData.pathCost[i][cond]*realTargetPriors[cond][allTargets.get(tf)];
				}
				else {
					assert Math.abs(pathData.pathCost[i][cond]) < 0.0001 : pathData.pathCost[i][cond];
					curIterPathCost[i] = 0;
				}
			}

			Arrays.sort(sortedPaths, new CompCost(curIterPathCost));
	
			double [] topPathsScore = nodePathPercent;
			double [] allPathsScore = new double[numNodes];
			double totalTopPathScore = 0.;
			double totalAllPathScore = 0.;
			int thresh = Math.min(pathThreshold, sortedPaths.length);
			int nPathsProcessed = 0;
			for (int j=0; j<thresh && nPathsProcessed<sortedPaths.length; ++nPathsProcessed) {
				int pathId = sortedPaths[nPathsProcessed];
				if (!pathData.pathActive[pathId]) {
					continue;
				}
				++j;
				int gene = networkData.edgeToNodes[paths[pathId][0]][0];
				topPathsScore[gene] += curIterPathCost[pathId];
				allPathsScore[gene] += curIterPathCost[pathId];
				for (int k=0; k<paths[pathId].length; ++k) {
					gene = networkData.edgeToNodes[paths[pathId][k]][1];
					topPathsScore[gene] += curIterPathCost[pathId];
					allPathsScore[gene] += curIterPathCost[pathId];
				}
				totalTopPathScore += curIterPathCost[pathId];
				totalAllPathScore += curIterPathCost[pathId];
			}
			
			for (int j=nPathsProcessed; j<sortedPaths.length; ++j) {
				int pathId = sortedPaths[j];
				if (!pathData.pathActive[pathId])
					continue;
				int gene = networkData.edgeToNodes[paths[pathId][0]][0];
				allPathsScore[gene] += curIterPathCost[pathId];
				for (int k=0; k<paths[pathId].length; ++k) {
					gene = networkData.edgeToNodes[paths[pathId][k]][1];
					allPathsScore[gene] += curIterPathCost[pathId];
				}
				totalAllPathScore += curIterPathCost[pathId];
			}
			
			for (int i=0; i<topPathsScore.length; ++i) {
				topPathsScore[i] /= totalTopPathScore;
				allPathsScore[i] /= totalAllPathScore;
			}
			
			if(!params.modelDir.equals("")) {
				try {
					// Write node scores
					String out = params.modelDir + "nodeScores_itr" + sdremIteration + 
							"_" + rankingMetric + "_" 
							+ params.orientationRestarts + "_" + pathThreshold + "_cond_" + cond+".txt";
					PrintWriter writer = new PrintWriter(new FileWriter(out));
					writer.println("ORF name\tStandard name\tSource\tTarget\t" + 
							"Weighted % top " + pathThreshold + " paths through node\t" +
							"Weighted % all paths through node");
	
					for (int j=0; j<numNodes; ++j) {
						assert !networkData.geneIdToName[j].trim().equals("");
						StringBuffer buf = new StringBuffer(
								networkData.geneIdToName[j] + "\t" + 
										networkData.geneIdToName[j]);
						if (networkData.sourceSet[cond].contains(j)) {
							buf.append("\t Y");
						}
						else {
							buf.append("\t N");
						}
						if (allTargets.getSet().contains(j) && realTargets[cond].contains(allTargets.get(j))) {
							buf.append("\t Y");
						}
						else {
							buf.append("\t N");
						}
						buf.append("\t" + topPathsScore[j] + "\t" + allPathsScore[j]);
						writer.println(buf.toString());
					}
					writer.close();

					int j=0;
					int num = 0;
					
					TIntHashSet selectedPaths = new TIntHashSet();
					for (j=0; j<pathData.pathActive.length; ++j) {
						if (!pathData.pathActive[j])
							continue;
						selectedPaths.add(j);
					}
					for (j=0; j<1000; ++j) {
						if (selectedPaths.contains(sortedPaths[j]))
							++num;
					}
					
					System.out.println("Num Paths in top 1,000 for cond " + cond + ": " + num);
					
					if (sortedPaths.length >= 10000) {
						for (; j<10000; ++j) {
							if (selectedPaths.contains(sortedPaths[j]))
								++num;
						}
						System.out.println("Num Paths in top 10,000: " + num);
					}

					if (sortedPaths.length >= 100000) {
						for (; j<100000; ++j) {
							if (selectedPaths.contains(sortedPaths[j]))
								++num;
						}
						System.out.println("Num Paths in top 100,000: " + num);
					}
				}
				catch (Exception e) {
					e.printStackTrace();
				}
			}
		}
	}
	
	public static void adjustPriorsGrid(
			int itr,
			ArraySet [] realTargets,
			double [][] realTargetPercentile,
			NetworkData networkData,
			Params params)
	{
		for (int i=0; i<params.numConditions; ++i) {
			String oldPrior = params.modelDir + params.tfGenePriorsFile + (itr - 1) + "_cond_" + i + ".txt";
			String newPrior = params.modelDir + params.tfGenePriorsFile + itr + "_cond_" + i + ".txt";
			try {
				BufferedReader reader = new BufferedReader(new FileReader(oldPrior));

				// The first row contains the TF names (except the first column)
				String line = reader.readLine();
				String tfs[] = line.split("\t");

				// Create arrays to determine how to modify the bound genes' prior
				// for each TF
				int incrTargCount = 0;
				int decrTargCount = 0;
				boolean[] change = new boolean[tfs.length];
				boolean[] increase = new boolean[tfs.length];
				if (tfs.length != networkData.numAllTargets) {
					throw new IllegalStateException(tfs.length + " != " + networkData.numAllTargets);
				}

				for(int t = 0; t < tfs.length; t++)
				{
					int tfIndex = networkData.allTargets.get(networkData.geneNameToId.get(tfs[t]));
					if(realTargets[i].contains(tfIndex))
					{
						change[t] = true;
	
						if(realTargetPercentile[i][tfIndex] >= params.targetThresh)
						{
							increase[t] = true;
							incrTargCount++;
						}
						else
						{
							increase[t] = false;
							decrTargCount++;
						}
					}
					else
					{
						change[t] = false;
						increase[t] = false;
					}
	
				}
				System.out.println("Increase prior of " + incrTargCount +
				" TFs because of target scores");
				System.out.println("Decrease prior of " + decrTargCount +
				" TFs because of target scores");
				
				int incrNodeCount = 0;
				for (int t=0; t<tfs.length; ++t) {
					int tfIndex = networkData.geneNameToId.get(tfs[t]);
					if (networkData.nodePriors[tfIndex] >= params.nodeThresh) {
						change[t] = true;
						increase[t] = true;
						incrNodeCount++;
					}
				}

				System.out.println("Increase prior of " + incrNodeCount
						+ " TFs because of node scores");

				// Write out the header again and then process each line individually
				// using the change and increase arrays
				PrintWriter writer = new PrintWriter(new FileWriter(newPrior));
				writer.println(line);

				while((line = reader.readLine()) != null) {
					String[] parts = line.split("\t");
					if(parts.length != tfs.length) {
						throw new IllegalStateException("Row beginning with " + parts[0] + 
								" does not contain " + tfs.length + " columns");
					}

					StringBuffer buf = new StringBuffer(parts[0]);

					for(int t = 1; t < tfs.length; t++) {
						buf.append("\t");
						if(change[t]) {
							if(Math.abs(Double.parseDouble(parts[t])) < 0.0001) {
								// Write 0 again if the TF doesn't bind this gene
								buf.append(parts[t]);
							}
							else {
								double val = Double.parseDouble(parts[t]);
								if(increase[t])
								{
									// The new value is the average of val and 1
									buf.append((1 + val) / 2.0);
								}
								else
								{
									// The new value is the average of val and 0
									// unless it is less than the minimum value
									buf.append(Math.max(val / 2.0, params.minPrior));
								}
							}
						}
						else {
							// Write the original value if this TF wasn't a target
							buf.append(parts[t]);
						}
					} // done processing a line

					writer.println(buf.toString());
				} // done reading file

				writer.close();
			}
			catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

	public static void adjustPriorsThreeCol(
			int itr,
			ArraySet [] realTargets,
			double [][] realTargetPercentile,
			double [] nodePathPercent,
			NetworkData networkData,
			Params params)
	{
		for (int i=0; i<params.numConditions; ++i) {
			String oldPrior = params.modelDir + params.tfGenePriorsFile + (itr - 1) + "_cond_" + i + ".txt";
			String newPrior = params.modelDir + params.tfGenePriorsFile + itr + "_cond_" + i + ".txt";
			final HashMap<String, Integer> geneNameToId = networkData.geneNameToId;
			final ArrayMapInt allTargets = networkData.allTargets;
			try {
				// Create arrays to determine how to modify the bound genes' prior
				// for each TF
				int incrTargCount = 0;
				int decrTargCount = 0;
				boolean[] change = new boolean[networkData.numAllTargets];
				boolean[] increase = new boolean[networkData.numAllTargets];

				for(int t = 0; t < networkData.numAllTargets; t++)
				{
					if(realTargets[i].contains(t)) {
						change[t] = true;
	
						if(realTargetPercentile[i][t] >= params.targetThresh) {
							increase[t] = true;
							incrTargCount++;
						}
						else {
							increase[t] = false;
							decrTargCount++;
						}
					}
					else {
						change[t] = false;
						increase[t] = false;
					}
				}

				System.out.println("Increase prior of " + incrTargCount +
				" TFs because of target scores");
				System.out.println("Decrease prior of " + decrTargCount +
				" TFs because of target scores");
				
				int incrNodeCount = 0;
				assert allTargets.getSet().size() == networkData.numAllTargets : allTargets.getSet().size() + " == " + networkData.numAllTargets;
				for (int t=0; t<networkData.numAllTargets; ++t) {
					if (realTargets[i].contains(t)) {
						continue;
					}
					if (nodePathPercent[networkData.targetIndexToTarget[t]] > params.nodeThresh) {
						change[t] = true;
						increase[t] = true;
						incrNodeCount++;
					}
				}

				System.out.println("Increase prior of " + incrNodeCount
						+ " TFs because of node scores");

				BufferedReader reader = new BufferedReader(new FileReader(oldPrior));

				// The first row contains the header
				String line = reader.readLine();

				// Write out the header again and then process each line individually
				// using the change and increase arrays
				PrintWriter writer = new PrintWriter(new FileWriter(newPrior));
				writer.println(line);

				while((line = reader.readLine()) != null) {
					String[] parts = line.split("\t");
					StringBuffer buf = new StringBuffer(parts[0]);
					
					if (!geneNameToId.containsKey(parts[0]) || !allTargets.getSet().contains(geneNameToId.get(parts[0]))) {
						writer.println(line);
						continue;
					}

					int tfId = allTargets.get(geneNameToId.get(parts[0]));
					
					if (change[tfId]) {
						double tempval;
						if (parts[2].equals("D")) {
							tempval = params.defaultTFprior;
						}
						else {
							tempval = Double.parseDouble(parts[2]);
						}
						
						if (Math.abs(tempval) < 0.0001) {
							writer.println(line);
						}
						else {
							if (increase[tfId]) {
								tempval = (1.0 + tempval)/2.0;
							}
							else {
								tempval = Math.max(tempval/2.0, params.minPrior);
							}
							
							writer.println(parts[0] + "\t" + parts[1] + "\t" + tempval);
						}
					}
					else {
						writer.println(line);
					}
				} // done reading file

				writer.close();
			}
			catch (Exception e) {
				e.printStackTrace();
			}
		}
	}
	
	public static class PathData {
		double [] curIterPathCost;
		int[][][] edgeToPaths;
		int[][] paths;
		int[][] undirPaths;
		double[][] pathCost;
		boolean[] pathActive;
	}
	
	public static class FlowData {
		double[] edgeFlow;
		double[] nodeFlow;
		double[] targetFlow;
	}


	public static void runSDREM(String propFile) throws IOException {
		NetworkData networkData = new NetworkData();
		Properties props = null;
		try {
			Properties defaults = new Properties();
			setDefaults(defaults);
	
			props = new Properties(defaults);
	
			// Load the properties from the specified file
			FileInputStream propsIn = new FileInputStream(propFile);
			props.load(propsIn);
			propsIn.close();
		}
		catch(Exception e) {
			e.printStackTrace();
		}
		
		Params params = new Params(props);

		readData(params, networkData);

		// Edge indexing starts at 1
		assert networkData.edgeToNodes.length == networkData.numEdges*2;
		double[] flipCost = new double[networkData.numEdges];
		byte[] edgeDir = new byte[networkData.edgeToNodes.length];
		
		Global.logger.info("Running SDREM");
		int firstItr = params.startIteration;

		if(!params.modelDir.endsWith("/"))
		{
			params.modelDir = params.modelDir + "/";
		}

		String metric = "PathWeight";
		
		Random r = new Random(params.seed);
		for (int sdremIter=firstItr; sdremIter<=params.iterations; ++sdremIter) {
			FileOutputStream fos = null;
			try {
				fos = new FileOutputStream(params.modelDir + "itr" + sdremIter + ".out");
			}
			catch (Exception e) {
				e.printStackTrace();
			}
			PrintStream ps = new PrintStream(fos);
			System.setOut(ps);

			if (sdremIter != firstItr || !params.useWarmDremStart) {
				for (int cI=0; cI<params.numConditions; ++cI) {
					String dremDefaults = params.modelDir + params.dremSettingsFile + "_cond_" + cI + ".txt";
					String oldPrior = params.modelDir + params.tfGenePriorsFile + (sdremIter - 1) + "_cond_" + cI + ".txt";
					String dremOut = params.modelDir + sdremIter + "_cond_" + cI;
					Global.logger.info("Starting DREM with config file " + dremDefaults);
					// Run DREM to generate the new activity scores
					DREMInterface.runDREM(dremDefaults, oldPrior, dremOut, params.dremRandomizations, params.distTfs, params.distThresh, cI);
		
					Calendar calendar = Calendar.getInstance();
					Date now = calendar.getTime();
					System.out.println(now);
					System.out.println("DREM finished for condition " + cI + "\n\n");
				}
			}

			double [][] realTargetPercentile = new double[params.numConditions][networkData.numAllTargets];
			ArraySet[] realTargets = new ArraySet[params.numConditions];
			for (int i=0; i<params.numConditions; ++i) {
				realTargets[i] = new ArraySet(networkData.numAllTargets);
			}
			double [][] realTargetPriors = new double[params.numConditions][networkData.numAllTargets];
			double[] nodePathPercent = new double[networkData.nodePriors.length];
			String dremTargetsFile = params.modelDir + sdremIter;
			readTargets(dremTargetsFile, realTargets, realTargetPriors, networkData, params);
			
			// Does edge orientation and gets the final TF priors
			
			if (params.networkMethod.equals("path")) {
				if (params.orientationRestarts > 0) {
					computeTfPriorsPathEnum(
							edgeDir, 
							flipCost, 
							networkData, 
							params, 
							sdremIter, 
							metric, 
							r,
							realTargets,
							realTargetPriors,
							realTargetPercentile);
				}

				rankedPathScorePathEnum(
						edgeDir,
						flipCost,
						realTargetPriors,
						sdremIter, 
						metric, 
						realTargets,
						nodePathPercent,
						networkData,
						params,
						r);
			}
			else if (params.networkMethod.equals("maxflow")) {
				if (params.orientationRestarts > 0) {
					computeTfPriorsMaxFlow(
							edgeDir, 
							flipCost, 
							networkData, 
							params, 
							sdremIter, 
							metric, 
							r,
							realTargets,
							realTargetPriors,
							realTargetPercentile);
				}

				rankedPathScoreMaxFlow(
						edgeDir,
						flipCost,
						realTargetPriors,
						sdremIter, 
						metric, 
						realTargets,
						nodePathPercent,
						networkData,
						params,
						r);
			}
			else {
				throw new IOException("Wrong method " + params.networkMethod + " given");
			}


			if (params.orientationRestarts > 0) {
				adjustPriorsThreeCol(sdremIter, realTargets, realTargetPercentile, nodePathPercent, networkData, params);
			}
		}
	}
	
	public static void readEdges(
			HashMap<String, Integer> geneNameToId,
			NetworkData networkData,
			final Params params)
	{
		try {
			ArrayList<ArrayList<Integer>> edgesTemp = new ArrayList<ArrayList<Integer>>();
			ArrayList<ArrayList<Byte>> edgesTypeTemp = new ArrayList<ArrayList<Byte>>();
			ArrayList<ArrayList<Double>> edgeCostTemp = new ArrayList<ArrayList<Double>>();
			BufferedReader edgeFileBr = new BufferedReader(new FileReader(params.edgeFile));
			String line = edgeFileBr.readLine();
			
			int numEdges = 0;
			for (int lineNum = 1; line != null; ++lineNum) {
				String [] splitLine = line.split("\t");
				if (splitLine.length != 4) {
					throw new IOException("Line in edge file doesn't have 4 fields");
				}
				
				for (int i=0; i<splitLine.length; ++i) {
					splitLine[i] = splitLine[i].trim();
				}
				
				if (!splitLine[1].equals("(pp)") && !splitLine[1].equals("(pd)") && !splitLine[1].equals("(ptm)")) {
					throw new IOException("Field 1 has illegal type edge " + splitLine[1] + " on line " + lineNum);
				}
				
				if (splitLine[0].equals("")) {
					throw new IOException("Field 0 has empty gene name on line " + lineNum);
				}
				if (splitLine[2].equals("")) {
					throw new IOException("Field 2 has empty gene name on line " + lineNum);
				}
				
				if (!geneNameToId.containsKey(splitLine[0])) {
					geneNameToId.put(splitLine[0], edgesTemp.size());
					edgesTemp.add(new ArrayList<Integer>());
					edgeCostTemp.add(new ArrayList<Double>());
					edgesTypeTemp.add(new ArrayList<Byte>());
				}
				if (!geneNameToId.containsKey(splitLine[2])) {
					geneNameToId.put(splitLine[2], edgesTemp.size());
					edgesTemp.add(new ArrayList<Integer>());
					edgeCostTemp.add(new ArrayList<Double>());
					edgesTypeTemp.add(new ArrayList<Byte>());
				}
				
				edgesTemp.get(geneNameToId.get(splitLine[0])).add(geneNameToId.get(splitLine[2]));
				edgeCostTemp.get(geneNameToId.get(splitLine[0])).add(Double.parseDouble(splitLine[3]));
				++numEdges;

				if (splitLine[1].equals("(pd)")) {
					edgesTypeTemp.get(geneNameToId.get(splitLine[0])).add((byte)2);
				}
				else if (splitLine[1].equals("(ptm)")) {
					edgesTypeTemp.get(geneNameToId.get(splitLine[0])).add((byte)1);
				}
				else if (splitLine[1].equals("(pp)")) {
					edgesTypeTemp.get(geneNameToId.get(splitLine[0])).add((byte)0);
				}

				line = edgeFileBr.readLine();
			}
			edgeFileBr.close();

			// Read TF names to put in gene list
			File folder = new File(params.storedPathsDir);
			File[] pathFiles = folder.listFiles();

			for (int i=0; i<pathFiles.length; ++i) {
				String pathFilename = pathFiles[i].getName();
				if (!pathFilename.endsWith(".gz")) {
					continue;
				}
				String [] pathFilenameSplit = pathFilename.split("\\.");
				String tf = pathFilenameSplit[0].trim();
				if (tf.equals("")) {
					throw new IOException("Empty tf name for filename " + pathFilename);
				}
				if (!geneNameToId.containsKey(tf)) {
					geneNameToId.put(tf, edgesTemp.size());
					edgesTemp.add(new ArrayList<Integer>());
					edgeCostTemp.add(new ArrayList<Double>());
					edgesTypeTemp.add(new ArrayList<Byte>());
				}
			}

			
			networkData.numEdges = numEdges;
			assert numEdges > 0;
			networkData.edges = new int[edgesTemp.size()][];
			networkData.edgeToId = new TIntIntHashMap[edgesTemp.size()];
			networkData.edgeToNodes = new int[2*numEdges][2];
			networkData.edgeType = new byte[numEdges];
			networkData.edgeCost = new double[2*numEdges];
			int [][] edges = networkData.edges;

			for (int i=0, k=0; i<edges.length; ++i) {
				edges[i] = new int[edgesTemp.get(i).size()];
				networkData.edgeToId[i] = new TIntIntHashMap();
				for (int j=0; j<edges[i].length; ++j, ++k) {
					edges[i][j] = edgesTemp.get(i).get(j);
					networkData.edgeToId[i].put(edges[i][j], k);
					networkData.edgeToNodes[k][0] = i;
					networkData.edgeToNodes[k][1] = edges[i][j];
					networkData.edgeToNodes[k+numEdges][0] = edges[i][j];
					networkData.edgeToNodes[k+numEdges][1] = i;
					networkData.edgeType[k] = edgesTypeTemp.get(i).get(j);
					networkData.edgeCost[k] = edgeCostTemp.get(i).get(j);
					networkData.edgeCost[k+numEdges] = edgeCostTemp.get(i).get(j);
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	public static void readNodePriors(
			final HashMap<String, Integer> geneNameToId,
			NetworkData networkData,
			final Params params)
	{
		double [] nodePriors = networkData.nodePriors;
		try {
			BufferedReader nodeFileBr = new BufferedReader(new FileReader(params.nodePriorsFile));
			String line = nodeFileBr.readLine();
			int num_priors_not_found = 0;

			while((line = nodeFileBr.readLine()) != null) {
				String [] splitLine = line.split("\t");
				if (splitLine.length != 2) {
					throw new IOException("Line has " + splitLine.length + " fields instead of 2");
				}
				if (geneNameToId.containsKey(splitLine[0])) {
					nodePriors[geneNameToId.get(splitLine[0])] = Double.parseDouble(splitLine[1]);
				}
				else {
					++num_priors_not_found;
				}
			}
			nodeFileBr.close();
			System.out.println(num_priors_not_found + " node prior genes not found");
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	public static void readSources(
			TIntHashSet [] sourceSet,
			final HashMap<String, Integer> geneNameToId,
			final Params params) throws Exception
	{
		for (int i=0; i<params.srcFiles.length; ++i) {
			BufferedReader reader = new BufferedReader(new FileReader(
					params.dataDir+params.srcFiles[i]));
			String line;
			int count = 0;
	
			while((line = reader.readLine()) != null)
			{
				String source = line.trim();
				if (geneNameToId.containsKey(source))
					sourceSet[i].add(geneNameToId.get(source));
				count++;
			}
	
			System.out.println("Loaded " + count + " sources for condition " + i);
			reader.close();
		}
	}
	
	public static int countPathsFromFiles(
			File[] pathFiles,
			TIntHashSet targets,
			NetworkData networkData,
			Params params)
	{
		int pathCount = 0;
		HashMap<String, HashSet<String>> sourceTfPairs = new HashMap<String, HashSet<String>>();
		try {
			for (int i=0; i<pathFiles.length; ++i) {
				String pathFilename = pathFiles[i].getName();
	
				if (!pathFilename.endsWith(".gz")) {
					continue;
				}
	
				String [] pathFilenameSplit = pathFilename.split("\\.");
				String tf = pathFilenameSplit[0];
	
				assert networkData.geneNameToId.containsKey(tf) : tf;
				int tfId = networkData.geneNameToId.get(tf);
				if (!targets.contains(tfId))
					continue;
				
	
				BufferedReader reader = new BufferedReader(new InputStreamReader(
						new GZIPInputStream(new FileInputStream(params.storedPathsDir+"/"+pathFilename))));
				String line;
				HashSet<String> sources = new HashSet<String>();
				while((line = reader.readLine()) != null) {
					String [] stringPathCost = line.split("\t");
					String [] pathSplit = stringPathCost[0].split("\\|");
	
					String source = pathSplit[0].split(":")[0];

	
					if (!sourceTfPairs.containsKey(source)
							|| !sourceTfPairs.get(source).contains(tf))
					{
						sources.add(source);
						++pathCount;
					}
				}
				reader.close();
	
				for (String s : sources) {
					if (!sourceTfPairs.containsKey(s)) {
						sourceTfPairs.put(s, new HashSet<String>());
					}
					sourceTfPairs.get(s).add(tf);
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		return pathCount;
	}
	
	public static class StringPath implements Comparable<StringPath> {
		String path;
		double weight;
		int id;
		
		public StringPath(String path, double weight, int id) {
			this.path = path;
			this.weight = weight;
			this.id = id;
		}

		public int compareTo(StringPath other) {
			int weightComp = Double.compare(weight, other.weight);
			if(weightComp == 0) {
				return path.compareTo(other.path);
			}
			else {
				return weightComp;
			}
		}
	}

	public static TIntHashSet getSortedPathsHeap(
			TIntHashSet targets,
			double[] targetPriors,
			int numPaths,
			File[] pathFiles,
			NetworkData networkData,
			Params params)
	{
		int pathCount = countPathsFromFiles(pathFiles, targets, networkData, params);
		Integer[] paths = new Integer[pathCount];
		for (int i=0; i<pathCount; ++i) {
			paths[i] = i;
		}

		HashMap<String, HashSet<String>> sourceTfPairs = new HashMap<String, HashSet<String>>();
		HashMap<String, Integer> geneNameToId = networkData.geneNameToId;
		PriorityQueue<StringPath> topPaths = new PriorityQueue<StringPath>();
		int k=0;
		try {
			for (int i=0; i<pathFiles.length; ++i) {
				String pathFilename = pathFiles[i].getName();
	
				if (!pathFilename.endsWith(".gz")) {
					continue;
				}
	
				String [] pathFilenameSplit = pathFilename.split("\\.");
				String tf = pathFilenameSplit[0];
				assert geneNameToId.containsKey(tf) : tf;
				int tfId = geneNameToId.get(tf);
				int tfIndex = networkData.allTargets.get(tfId);
				
				if (!targets.contains(tfId))
					continue;
				
				BufferedReader reader = new BufferedReader(new InputStreamReader(
						new GZIPInputStream(new FileInputStream(params.storedPathsDir+"/"+pathFilename))));
				String line;
				HashSet<String> sources = new HashSet<String>();
				while((line = reader.readLine()) != null) {
					String [] stringPathCost = line.split("\t");
					String [] pathSplit = stringPathCost[0].split("\\|");
					String source = pathSplit[0].split(":")[0];

					if (sourceTfPairs.containsKey(source)
							&& sourceTfPairs.get(source).contains(tf))
					{
						continue;
					}

					sources.add(source);
					StringPath newPath = new StringPath(
							stringPathCost[0], 
							Double.parseDouble(stringPathCost[1])*targetPriors[tfIndex],
							k++);

					if(topPaths.size() < numPaths) {
						topPaths.add(newPath);
					}
					else {
						StringPath lowPath = topPaths.peek();
						if(lowPath.compareTo(newPath) < 0)
						{
							topPaths.poll();
							topPaths.add(newPath);
						}
					}
				}
				reader.close();
	
				for (String s : sources) {
					if (!sourceTfPairs.containsKey(s)) {
						sourceTfPairs.put(s, new HashSet<String>());
					}
					sourceTfPairs.get(s).add(tf);
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		assert pathCount == k : pathCount + "\t" + k;
		
		TIntHashSet topPathSet = new TIntHashSet();
		for (StringPath sp : topPaths) {
			topPathSet.add(sp.id);
		}
		return topPathSet;
	}
	
	public static TIntHashSet getSortedPaths(
			TIntHashSet targets,
			double[] targetPriors,
			int numPaths,
			File[] pathFiles,
			NetworkData networkData,
			Params params)
	{
		int pathCount = countPathsFromFiles(pathFiles, targets, networkData, params);
		numPaths = Math.min(numPaths, pathCount);
		Integer[] paths = new Integer[pathCount];
		for (int i=0; i<pathCount; ++i) {
			paths[i] = i;
		}
		double [] cost = new double[pathCount];

		HashMap<String, HashSet<String>> sourceTfPairs = new HashMap<String, HashSet<String>>();
		HashMap<String, Integer> geneNameToId = networkData.geneNameToId;
		int k=0;
		try {
			for (int i=0; i<pathFiles.length; ++i) {
				String pathFilename = pathFiles[i].getName();
	
				if (!pathFilename.endsWith(".gz")) {
					continue;
				}
	
				String [] pathFilenameSplit = pathFilename.split("\\.");
				String tf = pathFilenameSplit[0];
				assert geneNameToId.containsKey(tf) : tf;
				int tfId = geneNameToId.get(tf);
				int tfIndex = networkData.allTargets.get(tfId);
				
				if (!targets.contains(tfId))
					continue;
				
				BufferedReader reader = new BufferedReader(new InputStreamReader(
						new GZIPInputStream(new FileInputStream(params.storedPathsDir+"/"+pathFilename))));
				String line;
				HashSet<String> sources = new HashSet<String>();
				while((line = reader.readLine()) != null) {
					String [] stringPathCost = line.split("\t");
					String [] pathSplit = stringPathCost[0].split("\\|");
	
					String source = pathSplit[0].split(":")[0];
	
					if (!sourceTfPairs.containsKey(source)
							|| !sourceTfPairs.get(source).contains(tf))
					{
						sources.add(source);
						cost[k++] = Double.parseDouble(stringPathCost[1])*targetPriors[tfIndex];
					}
				}
				reader.close();
	
				for (String s : sources) {
					if (!sourceTfPairs.containsKey(s)) {
						sourceTfPairs.put(s, new HashSet<String>());
					}
					sourceTfPairs.get(s).add(tf);
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		assert pathCount == k : pathCount + "\t" + k;

		Arrays.sort(paths, new CompCost(cost));
		
		TIntHashSet topPaths = new TIntHashSet();
		for (int i=0; i<numPaths; ++i) {
			topPaths.add(paths[i]);
		}
		return topPaths;
	}
	
	public static void readTargetsAndPaths(
			TIntHashSet targets,
			double[] targetPriors,
			int maxPaths,
			PathData pathData,
			final int numNodes,
			NetworkData networkData,
			final Params params) throws Exception
	{
		if (maxPaths == -1) {
			readTargetsAndPathsAll(pathData, numNodes, networkData, params);
			return;
		}

		File folder = new File(params.storedPathsDir);
		File[] pathFiles = folder.listFiles();
		
		HashMap<String, Integer> geneNameToId = networkData.geneNameToId;

		final int[][] edgeToNodes = networkData.edgeToNodes;
		final int numEdges = networkData.numEdges;
		final int [] countNumPathsForEdge = new int[numEdges];
		int totalNumPathEdgeCount = 0;
		
		TIntHashSet topPaths = getSortedPaths(targets, targetPriors, maxPaths, pathFiles, networkData, params);
		maxPaths = Math.min(maxPaths, topPaths.size());
		assert maxPaths > 0;
		//TIntHashSet topPaths = getSortedPathsHeap(targets, targetPriors, maxPaths, pathFiles, networkData, params);
		
		System.out.println("Reading paths from file");
		pathData.paths = new int[maxPaths][];
		pathData.undirPaths = new int[pathData.paths.length][];

		HashMap<String, HashSet<String>> sourceTfPairs = new HashMap<String, HashSet<String>>();
		final TIntIntHashMap [] edgeToId = networkData.edgeToId;
		Runtime runtime = Runtime.getRuntime();
		int l=0;
		TIntHashSet edgesUsed = new TIntHashSet();
		HashMap<String, Double> tfsUsed = new HashMap<String, Double>();
		for (int i=0, k=0; i<pathFiles.length; ++i) {
			String pathFilename = pathFiles[i].getName();

			if (!pathFilename.endsWith(".gz")) {
				continue;
			}
			
			String [] pathFilenameSplit = pathFilename.split("\\.");
			String tf = pathFilenameSplit[0];
			
			assert geneNameToId.containsKey(tf) : tf;
			int tfId = geneNameToId.get(tf);
			
			if (!targets.contains(tfId))
				continue;

			BufferedReader reader = new BufferedReader(new InputStreamReader(
					new GZIPInputStream(new FileInputStream(params.storedPathsDir+"/"+pathFilename))));
			String line;
			HashSet<String> sources = new HashSet<String>();
			while((line = reader.readLine()) != null)
			{
				String [] stringPathCost = line.split("\t");
				String [] pathSplit = stringPathCost[0].split("\\|");
				String source = pathSplit[0].split(":")[0];

				if (sourceTfPairs.containsKey(source) && sourceTfPairs.get(source).contains(tf)) {
					continue;
				}

				sources.add(source);
				
				if (!topPaths.contains(k)) {
					++k;
					continue;
				}
				
				double cost = Double.parseDouble(stringPathCost[1]);
				
				if (!tfsUsed.containsKey(tf)) {
					tfsUsed.put(tf, cost);
				}
				else {
					tfsUsed.put(tf, tfsUsed.get(tf)+cost);
				}

				int [] path = new int[pathSplit.length];
				int [] undirPath = new int[pathSplit.length];
				for (int j=0; j<pathSplit.length; ++j) {
					String [] edge = pathSplit[j].split(":");
					assert edge != null;
					assert geneNameToId != null;
					int [] nodes = {geneNameToId.get(edge[0]), geneNameToId.get(edge[2])};
					int undirEdge;
					if (edgeToId[nodes[0]].containsKey(nodes[1])) {
						undirEdge = edgeToId[nodes[0]].get(nodes[1]);
						path[j] = undirEdge;
					}
					else {
						assert edgeToId[nodes[1]].containsKey(nodes[0]);
						undirEdge = edgeToId[nodes[1]].get(nodes[0]);
						assert networkData.edgeType[undirEdge] == 0;
						path[j] = undirEdge+numEdges;
					}
					undirPath[j] = undirEdge;
					edgesUsed.add(undirEdge);
				}
				
				assert networkData.allTargets.getSet().contains(networkData.edgeToNodes[path[path.length-1]][1]) : networkData.geneIdToName[networkData.edgeToNodes[path[path.length-1]][1]] + "\t" + pathFilename;
				
				pathData.undirPaths[l] = undirPath;
				pathData.paths[l++] = path;
				for (int j=0; j<undirPath.length; ++j) {
					++countNumPathsForEdge[undirPath[j]];
					++totalNumPathEdgeCount;
				}
				++k;
			}
			
			for (String s : sources) {
				if (!sourceTfPairs.containsKey(s)) {
					sourceTfPairs.put(s, new HashSet<String>());
				}
				sourceTfPairs.get(s).add(tf);
			}
			reader.close();
		}
		assert l == topPaths.size() : l + "\t" + topPaths.size();
		System.out.println("Done reading paths from file w/ totalNumPathEdgeCount=" + totalNumPathEdgeCount + " and " + edgesUsed.size() + " edges and " + tfsUsed.size() + " tfs");
		
		pathData.edgeToPaths = new int[numEdges][][];
		
		int [][] paths = pathData.paths;
		int [][] undirPaths = pathData.undirPaths;
		int [][][] edgeToPaths = pathData.edgeToPaths;
		
		for (int i=0; i<numEdges; ++i) {
			edgeToPaths[i] = new int[countNumPathsForEdge[i]][2];
		}

		for (int i=0; i<paths.length; ++i) {
			for (int j=0; j<paths[i].length; ++j) {
				int undirEdge = undirPaths[i][j];
				--countNumPathsForEdge[undirEdge];
				edgeToPaths[undirEdge][countNumPathsForEdge[undirEdge]][0] = i;
				edgeToPaths[undirEdge][countNumPathsForEdge[undirEdge]][1] = j;
			}
		}
		pathData.curIterPathCost = new double[paths.length];

		System.out.println("Done setting up path datastruct");
		
	}
	
	public static void readTargetsAndPathsAll(
			PathData pathData,
			final int numNodes,
			NetworkData networkData,
			final Params params) throws Exception
	{
		File folder = new File(params.storedPathsDir);
		File[] pathFiles = folder.listFiles();

		HashMap<String, Integer> geneNameToId = networkData.geneNameToId;

		final int[][] edgeToNodes = networkData.edgeToNodes;
		final int numEdges = networkData.numEdges;
		final int [] countNumPathsForEdge = new int[numEdges];
		int totalNumPathEdgeCount = 0;
		int pathCount = 0;

		{
			HashMap<String, HashSet<String>> sourceTfPairs = new HashMap<String, HashSet<String>>();
			for (int i=0; i<pathFiles.length; ++i) {
				String pathFilename = pathFiles[i].getName();
	
				if (!pathFilename.endsWith(".gz")) {
					continue;
				}
	
				String [] pathFilenameSplit = pathFilename.split("\\.");
				String tf = pathFilenameSplit[0];
	
	
				BufferedReader reader = new BufferedReader(new InputStreamReader(
						new GZIPInputStream(new FileInputStream(params.storedPathsDir+"/"+pathFilename))));
				String line;
				HashSet<String> sources = new HashSet<String>();
				while((line = reader.readLine()) != null) {
					String [] stringPathCost = line.split("\t");
					String [] pathSplit = stringPathCost[0].split("\\|");
					
					String source = pathSplit[0].split(":")[0];
	
					if (!sourceTfPairs.containsKey(source)
							|| !sourceTfPairs.get(source).contains(tf))
					{
						sources.add(source);
						++pathCount;
					}
				}
				reader.close();
				
				for (String s : sources) {
					if (!sourceTfPairs.containsKey(s)) {
						sourceTfPairs.put(s, new HashSet<String>());
					}
					sourceTfPairs.get(s).add(tf);
				}
			}
		}

		System.out.println("Reading paths from file");
		pathData.paths = new int[pathCount][];
		pathData.undirPaths = new int[pathData.paths.length][];

		HashMap<Integer, HashSet<Integer>> sourceTfPairs = new HashMap<Integer, HashSet<Integer>>();
		final TIntIntHashMap [] edgeToId = networkData.edgeToId;
		Runtime runtime = Runtime.getRuntime();
		for (int i=0, k=0; i<pathFiles.length; ++i) {
			String pathFilename = pathFiles[i].getName();

			if (!pathFilename.endsWith(".gz")) {
				continue;
			}
			
			String [] pathFilenameSplit = pathFilename.split("\\.");
			String tf = pathFilenameSplit[0];
			assert geneNameToId.containsKey(tf);
			assert geneNameToId.containsKey(tf) : tf;
			int tfId = geneNameToId.get(tf);

			BufferedReader reader = new BufferedReader(new InputStreamReader(
					new GZIPInputStream(new FileInputStream(params.storedPathsDir+"/"+pathFilename))));
			String line;
			HashSet<Integer> sources = new HashSet<Integer>();
			while((line = reader.readLine()) != null)
			{
				String [] stringPathCost = line.split("\t");
				String [] pathSplit = stringPathCost[0].split("\\|");
				int [] path = new int[pathSplit.length];
				int [] undirPath = new int[pathSplit.length];
				for (int j=0; j<pathSplit.length; ++j) {
					String [] edge = pathSplit[j].split(":");
					assert edge != null;
					assert geneNameToId != null;
					int [] nodes = {geneNameToId.get(edge[0]), geneNameToId.get(edge[2])};
					int undirEdge;
					if (edgeToId[nodes[0]].containsKey(nodes[1])) {
						undirEdge = edgeToId[nodes[0]].get(nodes[1]);
						path[j] = undirEdge;
					}
					else {
						assert edgeToId[nodes[1]].containsKey(nodes[0]);
						undirEdge = edgeToId[nodes[1]].get(nodes[0]);
						path[j] = undirEdge+numEdges;
					}
					undirPath[j] = undirEdge;
					++countNumPathsForEdge[undirEdge];
					++totalNumPathEdgeCount;
				}
				assert networkData.allTargets.getSet().contains(networkData.edgeToNodes[path[path.length-1]][1]) : networkData.geneIdToName[networkData.edgeToNodes[path[path.length-1]][1]] + "\t" + pathFilename;
				
				if (!sourceTfPairs.containsKey(edgeToNodes[path[0]][0])
						|| !sourceTfPairs.get(edgeToNodes[path[0]][0]).contains(tfId))
				{
					sources.add(edgeToNodes[path[0]][0]);
					pathData.undirPaths[k] = undirPath;
					pathData.paths[k++] = path;
				}
			}
			
			for (Integer s : sources) {
				if (!sourceTfPairs.containsKey(s)) {
					sourceTfPairs.put(s, new HashSet<Integer>());
				}
				sourceTfPairs.get(s).add(tfId);
			}
			reader.close();
		}
		System.out.println("Done reading paths from file w/ totalNumPathEdgeCount=" + totalNumPathEdgeCount);

		pathData.edgeToPaths = new int[numEdges][][];
		
		int [][] paths = pathData.paths;
		int [][] undirPaths = pathData.undirPaths;
		int [][][] edgeToPaths = pathData.edgeToPaths;
		
		for (int i=0; i<numEdges; ++i) {
			edgeToPaths[i] = new int[countNumPathsForEdge[i]][2];
		}

		for (int i=0; i<paths.length; ++i) {
			for (int j=0; j<paths[i].length; ++j) {
				int undirEdge = undirPaths[i][j];
				--countNumPathsForEdge[undirEdge];
				edgeToPaths[undirEdge][countNumPathsForEdge[undirEdge]][0] = i;
				edgeToPaths[undirEdge][countNumPathsForEdge[undirEdge]][1] = j;
			}
		}
		pathData.curIterPathCost = new double[paths.length];

		System.out.println("Done setting up path datastruct");
	}

	public static void readData(
			final Params params,
			NetworkData networkData) 
	{
		try {
			HashMap<String, Integer> geneNameToId = networkData.geneNameToId;
			
			readEdges(geneNameToId, networkData, params);
			
			networkData.geneIdToName = new String[geneNameToId.size()];
			for (Map.Entry<String, Integer> entry : geneNameToId.entrySet()) {
				assert !entry.getKey().trim().equals("");
				networkData.geneIdToName[entry.getValue()] = entry.getKey();
			}
			
			networkData.nodePriors = new double[networkData.edges.length];
			for (int i=0; i<networkData.edges.length; ++i) {
				networkData.nodePriors[i] = params.defaultNodePrior;
			}
			networkData.numNodes = networkData.nodePriors.length;
			
			if (!params.nodePriorsFile.equals("")) {
				readNodePriors(geneNameToId, networkData, params);
			}
			
			networkData.sourceSet = new TIntHashSet[params.srcFiles.length];
			for (int i=0; i<params.srcFiles.length; ++i) {
				networkData.sourceSet[i] = new TIntHashSet();
			}
			readSources(networkData.sourceSet, geneNameToId, params);

			networkData.allTargets = new ArrayMapInt(networkData.nodePriors.length);
			ArrayMapInt allTargets = networkData.allTargets;

			File folder = new File(params.storedPathsDir);
			File[] pathFiles = folder.listFiles();
			networkData.targetIndexToTarget = new int[pathFiles.length];
			for (int i=0, k=0; i<pathFiles.length; ++i) {
				String pathFilename = pathFiles[i].getName();
	
				if (!pathFilename.endsWith(".gz")) {
					continue;
				}
				
				String [] pathFilenameSplit = pathFilename.split("\\.");
				String tf = pathFilenameSplit[0];
				assert geneNameToId.containsKey(tf);
				assert geneNameToId.containsKey(tf) : tf;
				int tfId = geneNameToId.get(tf);
				allTargets.add(tfId, networkData.numAllTargets);
				networkData.targetIndexToTarget[networkData.numAllTargets++] = geneNameToId.get(tf);
			}
			assert networkData.numAllTargets == allTargets.getSet().size();
			System.out.println("Loaded " + networkData.numAllTargets + " targets");
		}
		catch(Exception e) {
			e.printStackTrace();
		}
	}
	
	public static void setDefaults(Properties defaults)
	{
		defaults.setProperty("edges.file", "example/sampleEdges.txt");
		defaults.setProperty("sources.file", "example/sampleSources.txt");
		defaults.setProperty("model.dir", "example");
		defaults.setProperty("drem.settings.file", "dremSettings.txt");
		defaults.setProperty("binding.priors.file", "bindingPriors");
		defaults.setProperty("node.priors.file", ""); // Do not load specific node priors
		defaults.setProperty("stored.paths.dir", ""); // Do not load precomputed paths
		defaults.setProperty("iterations", "10");
		defaults.setProperty("max.path.length", "5");
		defaults.setProperty("path.enum.bound", "-1"); // -1 Signifies that all paths should be enumerated
		defaults.setProperty("orientation.restarts", "20");
		defaults.setProperty("drem.random.runs", "10");
		defaults.setProperty("dist.tfs", "50");
		defaults.setProperty("top.paths", "-1"); // -1 Signifies 5 * number of targets
		defaults.setProperty("dist.thresh", "0.5");
		defaults.setProperty("target.thresh", "0.8");
		defaults.setProperty("node.thresh", "0.01");
		defaults.setProperty("default.node.prior", "1.0");
		defaults.setProperty("min.prior", "0.01");
		defaults.setProperty("random.target.ratio", "1");
		defaults.setProperty("startIteration", "1");
		defaults.setProperty("seed", "1");
		defaults.setProperty("num.conditions", "1");
		defaults.setProperty("default.TF.prior", "0.5");
		defaults.setProperty("synFile", "");
		defaults.setProperty("DEBUG", "true");
		defaults.setProperty("XDEBUG", "false");
		defaults.setProperty("opt.convergence.ratio", "0.1");
		defaults.setProperty("WarmDremStart", "false");
		defaults.setProperty("tabu.size", "0");
		defaults.setProperty("tabu.maxitr", "0");
		defaults.setProperty("cond.class", "");
		defaults.setProperty("tf.share.factor", "1.0");
		defaults.setProperty("maxflow.source.penalty", "0.0");
		defaults.setProperty("network.method", "path");
	}
}
