package util;

public class ArraySet {
	private int capacity;
	private int offset;
	private int size;
	private int[] list;
	private int[] marked;
	
	public ArraySet(int capacity) {
		this(capacity, 0, false);
	}
	
	public ArraySet(int capacity, int offset) {
		this(capacity, offset, false);
	}
	
	public ArraySet(int capacity, int offset, boolean initAll) {
		this.capacity = capacity;
		this.offset = offset;
		list = new int[capacity];
		marked = new int[capacity];
		
		if (initAll) {
			size = capacity;
			for (int i=0; i<capacity; ++i) {
				list[i] = i;
				marked[i] = i;
			}
		}
		else {
			for (int i=0; i<capacity; ++i) {
				marked[i] = -1;
			}
		}
	}
	
	public boolean add(int element) {
		assert (element >= offset) : element + " >= " + offset;
		element -= offset;
		assert (element < capacity && element >= 0) : element + " < " + capacity + " and >= 0";
		
		if (marked[element] != -1) {
			return false;
		}
		
		marked[element] = size;
		list[size++] = element;
		return true;
	}
	
	public int get(int pos) {
		assert (pos < size) : pos + " < " + size;
		return list[pos]+offset;
	}
	
	public boolean remove(int element) {
		assert (element >= offset) : element + " >= " + offset;
		element -= offset;
		assert (element < capacity && element >= 0) : element + " < " + capacity + " and >= 0";
		
		if (marked[element] == -1) {
			return false;
		}
		
		list[marked[element]] = list[--size];
		marked[list[size]] = marked[element];
		marked[element] = -1;
		return true;
	}
	
	public void clear() {
		for (int i=0; i<size; ++i) {
			marked[list[i]] = -1;
		}
		size = 0;
	}
	
	public boolean contains(int element) {
		assert (element >= offset) : element + " >= " + offset;
		element -= offset;
		assert (element < capacity && element >= 0) : element + " < " + capacity + " and >= 0";
		
		return marked[element] != -1;
	}
	
	public int size() {
		return this.size;
	}
	
	public int maxElement() {
		return offset+capacity-1;
	}
	
	public int minElement() {
		return offset;
	}

	public boolean isSubset(ArraySet s) {
		if (size() > s.size())
			return false;
		
		// If condition here for efficiency purposes - so the extra check
		// in the for loop can be avoided if the else condiiton holds.
		// This could be made more efficient if the min/max elements were
		// kept track of.
		int sMinElement = s.minElement();
		int sMaxElement = s.maxElement();
		if (minElement() < sMinElement || maxElement() > sMaxElement) {
			for (int i=0; i<size; ++i) {
				int el = get(i);
				if (el < sMinElement || el > sMaxElement)
					return false;
				if (!s.contains(get(i)))
					return false;
			}
		}
		else {
			for (int i=0; i<size; ++i) {
				if (!s.contains(get(i)))
					return false;
			}
		}
		return true;
	}
	
	public ArraySet setMinus(ArraySet b) {
		ArraySet result = new ArraySet(capacity, offset);
		setMinus(b, result);
		return result;
	}
	
	public void setMinus(ArraySet b, ArraySet result) {
		int sMinElement = b.minElement();
		int sMaxElement = b.maxElement();
		for (int i=0; i<size; ++i) {
			int el = get(i);
			if (el < sMinElement || el > sMaxElement || !b.contains(el)) {
				result.add(el);
			}
		}
	}
}
